# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Rules for coverage:

- new routes on init get count 0
- new routes after init get count 1 per context
- routes without a method type specified on init are ['GET', 'POST']
- a route needs verb, route, url, and count
- url should be normalized uri
- route should be the path to the view function (aka controller for Java people)
- one method type per route

Example:
      GET /blog/foo/bar - app.blogs.find(request, ) 0
"""

import functools
import inspect
from typing import Callable
import re

from contrast.utils.decorators import fail_quietly

from contrast_fireball import DiscoveredRoute
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


DEFAULT_ROUTE_METHODS = ("GET", "POST")
ROUTE_DISCOVERY_THREAD_NAME = "ContrastRouteDiscovery"

# A format template for constructing "path-component" regexps, ie between "/"-s in a URI
# This is used to build regexps for uri normalization
# We use a look-ahead to ensure correctness for back-to-back path components
_RE_PATH_COMPONENT_FORMAT = r"(^|/){regex}(?=(/|$))"

# examples:
# /123
# 123/
# 123
# /123/
# 123/123
RE_DIGITS_PATH_COMPONENT = _RE_PATH_COMPONENT_FORMAT.format(regex=r"\d+")

# example:
# /fd4b78312634a236d11da0f9c32526e5b8261afa
RE_HASH_PATH_COMPONENT = _RE_PATH_COMPONENT_FORMAT.format(
    regex=r"([a-fA-F0-9]{2}){16,}"
)

# example:
# /b09112a0-a58f-487a-ab4b-3608bd64fb3f/
RE_UUID_PATH_COMPONENT = _RE_PATH_COMPONENT_FORMAT.format(
    regex=r"[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}"
)

# Returns digits between '/' or end of string
FIND_REQEUST_PARAMS = r"(?<=\/)\d+(?=\/|$)"


def get_normalized_uri(path):
    """
    A best-effort to remove client-specific information from the path.

    Example:
    /user/123456/page/12 -> /user/{n}/page/{n}
    """
    result = path

    result = re.sub(RE_DIGITS_PATH_COMPONENT, r"\1{n}", result)
    result = re.sub(RE_HASH_PATH_COMPONENT, r"\1{hash}", result)
    result = re.sub(RE_UUID_PATH_COMPONENT, r"\1{uuid}", result)

    return result


def get_url_parameters(path):
    """
    A best-effort to remove client-specific information from the path.

    Example:
    /user/123456/page/12 -> /user/{n}/page/{n}
    """
    return re.findall(FIND_REQEUST_PARAMS, path)


def build_signature(view_func_name, view_func):
    view_func_args = build_args_from_function(view_func)
    return view_func_name + view_func_args


def build_args_from_function(func):
    """
    Attempts to grab argument names from the function definition.

    Defaults to () if none exist
    If there is no view function, like in the case of a pure WSGI app, then the func will
    be a string like '/sqli' and we just return that.

    """
    if isinstance(func, functools.partial):
        func = getattr(func, "func", None)

    func = inspect.unwrap(func)
    method_arg_names = "()"
    if func is not None and hasattr(func, "__code__"):
        method_arg_names = str(
            func.__code__.co_varnames[0 : func.__code__.co_argcount]
        ).replace("'", "")
    elif isinstance(func, str):
        method_arg_names = func

    return method_arg_names


@fail_quietly("Failed to perform route discovery")
def handle_route_discovery(
    framework: str, discovery_func: Callable[..., set[DiscoveredRoute]], args
) -> None:
    """
    Start the route discovery background thread if inventory is enabled.
    """
    from contrast.agent import agent_state

    if (
        settings := agent_state.get_settings()
    ) is None or not settings.is_inventory_enabled():
        logger.debug("Inventory disabled - will not perform route discovery")
        return

    logger.debug("Starting route discovery background thread", framework=framework)
    _do_route_discovery(framework, discovery_func, args)


@fail_quietly("Failed to perform route discovery in background thread")
def _do_route_discovery(
    framework: str,
    discovery_func: Callable[..., set[DiscoveredRoute]],
    args: tuple,
) -> None:
    """
    Top-level background thread function for route discovery.
    """
    from contrast.agent import agent_state

    discovered_routes = discovery_func(*args)
    agent_state.module.reporting_client.new_discovered_routes(discovered_routes)
    log_discovered_routes(framework, discovered_routes)


def log_discovered_routes(framework: str, routes: set[DiscoveredRoute]) -> None:
    logger.debug(
        "Discovered routes",
        framework=framework,
        routes=[f"{route.verb} {route.url}" for route in routes],
    )



from __future__ import annotations

from copy import copy
from dataclasses import dataclass, field
import json
import time

import pytest
from syrupy.assertion import SnapshotAssertion
from syrupy.matchers import path_type
from syrupy.types import PropertyMatcher, PropertyPath, SerializableData
from testcontainers.generic.server import ServerContainer
from conftest import local_path


@pytest.mark.parametrize("endpoint", ["/clock/", "/sleep/"])
def test_system_commands(
    observe_django_app: ServerContainer,
    snapshot_json: SnapshotAssertion,
    endpoint: str,
):
    client = observe_django_app.get_client()

    response = client.get(endpoint)
    assert response.status_code == 200, "Failed to connect to Django app"

    trace = find_trace_for_path(endpoint)
    assert_trace_snapshot(snapshot_json, trace)


def test_file_opens_write(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    post_name = "test_post.txt"
    post_content = "This is a test post."
    create_response = client.post(f"/posts/{post_name}/", content=post_content)
    assert create_response.status_code == 200

    trace = find_trace_for_path(f"/posts/{post_name}/")
    assert_trace_snapshot(snapshot_json, trace)


def test_file_opens_read(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    retrieve_response = client.get("/posts/example.txt/")
    assert retrieve_response.status_code == 200

    trace = find_trace_for_path("/posts/example.txt/")
    assert_trace_snapshot(snapshot_json, trace)


def test_file_opens_list(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    list_response = client.get("/posts/")
    assert list_response.status_code == 200

    # Listing a directoy doesn't open any files, so there shouldn't be
    # any file-open-create spans.
    trace = find_trace_for_path("/posts/")
    assert_trace_snapshot(snapshot_json, trace)


def test_storage_query(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    # Create a new message
    msg_text = "Hello, database!"
    create_response = client.post(
        "/messages/",
        content=json.dumps({"text": msg_text}),
        headers={"Content-Type": "application/json"},
    )
    assert create_response.status_code == 201, create_response.content

    # Query all messages
    list_response = client.get("/messages/")
    assert list_response.status_code == 200

    trace = find_trace_for_path("/messages/")
    assert_trace_snapshot(snapshot_json, trace)


def test_django_auth_login(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    # Prep the csrf token by visiting the login page first
    client.get("/admin/login/", follow_redirects=True)

    # Now log in, which actually performs the authentication request.
    client.post(
        "/admin/login/",
        data={
            "username": "admin",
            "password": "password",
            "csrfmiddlewaretoken": client.cookies["csrftoken"],
        },
        headers={"Content-Type": "application/x-www-form-urlencoded"},
        follow_redirects=True,
    )

    trace = find_trace_for_path("/admin/login/")
    assert_trace_snapshot(snapshot_json, trace)


def test_django_authn_session(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    # Currently, we don't distinguish between successful and unsuccessful authn requests.
    # So we keep it simple and don't worry about authenticating before making the test.
    response = client.get("/auth/test/")
    assert response.status_code == 401

    trace = find_trace_for_path("/auth/test/")
    assert_trace_snapshot(snapshot_json, trace)


def test_outbound_request(
    observe_django_app: ServerContainer, snapshot_json: SnapshotAssertion
):
    client = observe_django_app.get_client()

    response = client.get("/external-data/")
    assert response.status_code == 200
    assert response.json() == {"result": "mocked response for testing"}

    trace = find_trace_for_path("/external-data/")
    assert_trace_snapshot(snapshot_json, trace)


def assert_trace_snapshot(snapshot_json: SnapshotAssertion, trace: dict):
    assert (
        snapshot_json(
            matcher=multi_matcher(
                attributes_sorter,
                attribute_stabilizer("service.instance.id"),
                attribute_stabilizer("service.name"),  # includes agent version
                path_type(
                    {
                        r"resourceSpans.\d+.scopeSpans.\d+.spans.\d+.startTimeUnixNano": (
                            str,
                        ),
                        r"resourceSpans.\d+.scopeSpans.\d+.spans.\d+.endTimeUnixNano": (
                            str,
                        ),
                        r"resourceSpans.\d+.scopeSpans.\d+.scope.version": (str,),
                    },
                    regex=True,
                ),
                TraceSpanIDsMatcher(),
            )
        )
        == trace
    )


def attributes_sorter(data: SerializableData, path: PropertyPath) -> SerializableData:
    """
    A syrupy matcher that sorts the attributes dictionary by key.
    This is useful for ensuring consistent snapshots when the order of attributes
    may vary between runs.
    """
    if path and path[-1] == ("attributes", list):
        return list(sorted(data, key=lambda item: item["key"]))

    return data


def attribute_stabilizer(attr_name: str) -> PropertyMatcher:
    """
    Creates a syrupy matcher that replaces the named attribute if it matches the expected type.
    """

    def replacer(data: SerializableData, path: PropertyPath) -> SerializableData:
        if len(path) >= 2 and path[-1] == ("attributes", list):

            def replaced_data(key, value):
                if key == attr_name:
                    return {"key": key, "value": {"stringValue": f"<stable-{key}>"}}
                else:
                    return {"key": key, "value": value}

            return [replaced_data(kv["key"], kv["value"]) for kv in data]
        return data

    return replacer


@dataclass
class TraceSpanIDsMatcher:
    """
    A matcher that normalizes trace and span IDs in the data.

    OTel trace and span IDs are unique identifiers that aren't stable across runs.
    This matcher replaces them with a stable string representation based on their
    order of appearance in the data. This allows for consistent snapshots without
    losing the uniqueness of the IDs.

    If the order of spans or traces changes, the backend may still be able to render
    the trace correctly, but this matcher will fail. This is taking the simple approach
    first, and we can improve it later if needed.
    """

    trace_ids: dict[str, str] = field(default_factory=dict)
    span_ids: dict[str, str] = field(default_factory=dict)

    def __call__(self, data: SerializableData, path: PropertyPath) -> SerializableData:
        if not path:
            return data

        if path[-1] == ("traceId", str):
            bucket = self.trace_ids
        elif path[-1] in [("spanId", str), ("parentSpanId", str)]:
            bucket = self.span_ids
        else:
            return data

        if data in bucket:
            return bucket[data]
        elif path[-1] == ("parentSpanId", str) and data == "":
            return data  # Keep empty parentSpanId as is
        else:
            normalized_id = bucket[data] = str(len(bucket))
            return normalized_id


def multi_matcher(*matchers: PropertyMatcher) -> PropertyMatcher:
    """
    Combines multiple matchers into a single matcher.
    """

    def combined_replacer(
        data: SerializableData, path: PropertyPath
    ) -> SerializableData:
        for matcher in matchers:
            new_data = matcher(data=data, path=path)
            if new_data is not data:
                data = new_data
        return data

    return combined_replacer


def poll_traces(timeout=10, poll_interval=1):
    """
    Generator that yields the current traces every poll_interval seconds, up to timeout seconds.
    """
    start = time.time()
    while time.time() - start < timeout:
        traces = [
            load_json_from_file(log_file)
            for log_file in local_path("logs/messages/requests").glob(
                "*-observability-traces-teamserver.json"
            )
        ]
        yield traces
        time.sleep(poll_interval)


def find_trace_for_path(path: str, timeout: int = 10) -> dict:
    """
    Search for and return the OpenTelemetry trace containing all spans for a given HTTP path.

    This function scans the proxy logs for batch trace files, searching for a span whose "url.path"
    attribute matches the provided path. Once found, it filters the trace batch to include only the
    spans that match the specified path.
    """
    for traces in poll_traces(timeout=timeout):
        (batch_trace, request_trace_ids) = find_trace_id(traces, path)
        if batch_trace and request_trace_ids:
            matching_resource_spans = []
            for resource_span in batch_trace.get("resourceSpans", []):
                matching_scope_spans = []
                for scope_span in resource_span.get("scopeSpans", []):
                    matching_spans = [
                        span
                        for span in scope_span.get("spans", [])
                        if span.get("traceId") in request_trace_ids
                    ]
                    if matching_spans:
                        scope_span_copy = copy(scope_span)
                        scope_span_copy["spans"] = matching_spans
                        matching_scope_spans.append(scope_span_copy)
                if matching_scope_spans:
                    resource_span_copy = copy(resource_span)
                    resource_span_copy["scopeSpans"] = matching_scope_spans
                    matching_resource_spans.append(resource_span_copy)
            batch_trace_copy = copy(batch_trace)
            batch_trace_copy["resourceSpans"] = matching_resource_spans
            return batch_trace_copy

    raise TimeoutError(
        f"No observe trace requests found for path '{path}' within the {timeout} seconds."
    )


def find_trace_id(batch_traces, path):
    """
    Find all trace IDs for the given path in the traces.
    Returns the first matching batch_trace and a list of all matching trace IDs, or (None, None) if not found.
    """
    for batch_trace in batch_traces:
        trace_ids = []
        for resource_span in batch_trace.get("resourceSpans", []):
            for scope_span in resource_span.get("scopeSpans", []):
                for span in scope_span.get("spans", []):
                    for attr in span.get("attributes", []):
                        if (
                            attr.get("key") == "url.path"
                            and attr.get("value", {}).get("stringValue") == path
                        ):
                            trace_id = span["traceId"]
                            if trace_id not in trace_ids:
                                trace_ids.append(trace_id)
        if trace_ids:
            return batch_trace, trace_ids
    return None, None


def load_json_from_file(path):
    with open(path) as f:
        return json.load(f)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import (
    copy_events,
    copy_tags_to_offset,
    copy_from,
    get_last_event_from_sources,
    get_properties,
)

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class AppendPropagator(BasePropagator):
    def get_parents(self, ret):
        return get_last_event_from_sources(self.sources)

    def _propagate(self):
        second_source = None

        if len(self.sources) > 1:
            second_source = self.sources[1]

        # if the object and the return are the same length just copy the tags
        # from the object(since nothing from args was added to return)
        if len(self.first_source) == len(self.target):
            copy_from(self.target, self.first_source, 0, self.node.untags)
        else:
            # find original in the target, copy tags to the new position in target
            try:
                original_start_index = self.target.index(self.first_source)
            except Exception:
                logger.debug(
                    "%s was not found in the target %s", self.first_source, self.target
                )
                return

            copy_from(
                self.target, self.first_source, original_start_index, self.node.untags
            )

            start = original_start_index + len(self.first_source)
            if second_source is not None:
                source_properties = get_properties(second_source)
                if source_properties is not None:
                    copy_tags_to_offset(
                        self.target_properties, source_properties.tags, start
                    )
                    copy_events(self.target_properties, source_properties)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import json
import re

from contrast.agent.assess.rules.response.base_header_only_rule import (
    BaseResponseRule,
)
from contrast.agent.assess.rules import build_finding, send_finding
from contrast.utils.decorators import fail_quietly

UNSAFE_RE = re.compile(r"unsafe-(inline|eval)")

DEFAULT_SRC = "default-src"
CHILD_SRC = "child-src"
CONNECT_SRC = "connect-src"
FRAME_SRC = "frame-src"
MEDIA_SRC = "media-src"
OBJECT_SRC = "object-src"
SCRIPT_SRC = "script-src"
STYLE_SRC = "style-src"
BASE_URI = "base-uri"
FORM_ACTION = "form-action"
FRAME_ANCESTORS = "frame-ancestors"


class CspHeaderInsecureRule(BaseResponseRule):
    """
    Rule is violated if any the following headers:

    - Content-Security-Policy
    - X-Content-Security-Policy
    - X-Webkit-CSP

    are set with insecure parameters.

    Note that CSP-header-missing is a separate rule.
    """

    @property
    def name(self):
        return "csp-header-insecure"

    @property
    def header_keys(self):
        return ("Content-Security-Policy", "X-Content-Security-Policy", "X-Webkit-CSP")

    def is_violated(self, headers, _, form_tags, meta_tags):
        header_violated, header_properties = self.is_header_violated(headers)
        if not header_violated:
            # If header is correctly configured for this rule, no need to check body.
            return False, {}

        body_violated, body_properties = self.is_body_violated(meta_tags)
        if not body_violated:
            # Even though header may have violated the rule, if the body was configured
            # correctly for this rule, the rule is not violated.
            return False, {}

        return True, header_properties or body_properties

    def is_header_violated(self, headers):
        for key in self.header_keys:
            value = headers.get(key)
            if not value:
                continue

            csp_instructions = _detect_csp_settings(value)

            if csp_instructions is None or csp_instructions.is_safe():
                continue

            return True, csp_instructions.to_properties()

        return False, None

    def is_body_violated(self, meta_tags):
        """
        Rule is violated if:
        1. none of the attrs of the list of form tags have an "http-equiv" attr
        2. at least one attr is "http-equiv" and the next one in the current meta tag is not "content"
        :param form_tags: list of Tag namedtuple
        :return: violated bool, properties dict
        """
        if not meta_tags:
            return True, {}

        for tag in meta_tags:
            attrs_dict = {k.lower(): v.lower() for k, v in tag.attrs}
            if (
                attrs_dict.get("http-equiv") == "content-security-policy"
                and "content" in attrs_dict
            ):
                csp_instructions = _detect_csp_settings(attrs_dict.get("content"))
                if not csp_instructions.is_safe():
                    return True, csp_instructions.to_properties()

        return False, None

    def create_finding(self, properties):
        return build_finding(self, {"data": json.dumps(properties)})

    def build_and_append_finding(self, properties, context):
        finding = self.create_finding(properties)
        send_finding(finding, context)


@fail_quietly("Failed to detect CSP settings")
def _detect_csp_settings(csp_header_value):
    """
    Parse and detect safety of the entire CSP header. The format of CSP is a series of
    semi-colon separated directives, where each directive consists of a key followed by
    several values, all separated by whitespace. For example:

    "default-src 'self' cdn.example.com; script-src 'self'; connect-src 'self'"

    parses to:

    {
        "default-src": ["'self'", "cdn.example.com"],
        "script-src": ["'self'"],
        "connect-src": ["'self'"],
    }
    """
    split_values = [d.split(sep=None, maxsplit=1) for d in csp_header_value.split(";")]
    parsed_values = {v[0].lower(): v[1].lower() for v in split_values if len(v) > 1}
    return _CSPInstructions(parsed_values)


def _is_content_secure(value):
    return not (value is None or "*" in value)


def _is_content_safe(value):
    return not UNSAFE_RE.match(value)


class _CSPInstructions:
    def __init__(self, parsed_values):
        self.default_src = parsed_values.get(DEFAULT_SRC)
        self.child_src = parsed_values.get(CHILD_SRC)
        self.connect_src = parsed_values.get(CONNECT_SRC)
        self.frame_src = parsed_values.get(FRAME_SRC)
        self.media_src = parsed_values.get(MEDIA_SRC)
        self.object_src = parsed_values.get(OBJECT_SRC)
        self.script_src = parsed_values.get(SCRIPT_SRC)
        self.style_src = parsed_values.get(STYLE_SRC)
        self.base_uri = parsed_values.get(BASE_URI)
        self.form_action = parsed_values.get(FORM_ACTION)
        self.frame_ancestors = parsed_values.get(FRAME_ANCESTORS)

        self.default_src_secure = _is_content_secure(self.default_src)
        self.child_src_secure = _is_content_secure(self.child_src)
        self.connect_src_secure = _is_content_secure(self.connect_src)
        self.frame_src_secure = _is_content_secure(self.frame_src)
        self.media_src_secure = _is_content_secure(self.media_src)
        self.object_src_secure = _is_content_secure(self.object_src)
        self.script_src_secure = _is_content_secure(
            self.script_src
        ) and _is_content_safe(self.script_src)
        self.style_src_secure = _is_content_secure(self.style_src) and _is_content_safe(
            self.style_src
        )
        self.base_uri_secure = _is_content_secure(self.base_uri)
        self.form_action_secure = _is_content_secure(self.form_action)
        self.frame_ancestors_secure = _is_content_secure(self.frame_ancestors)

    def is_safe(self):
        return (
            self.form_action_secure
            and self.frame_ancestors_secure
            and self.base_uri_secure
            and self.are_sources_secure()
        )

    def are_sources_secure(self):
        extra_sources = [
            "child_src",
            "frame_src",
            "connect_src",
            "media_src",
            "object_src",
            "script_src",
            "style_src",
        ]
        if self.default_src_secure:
            for src in extra_sources:
                if getattr(self, src) is not None and not getattr(
                    self, f"{src}_secure"
                ):
                    return False
            return True

        return all([getattr(self, f"{src}_secure") for src in extra_sources])

    def to_properties(self):
        return {
            "defaultSrcValue": self.default_src or "",
            "childSrcValue": self.child_src or "",
            "connectSrcValue": self.connect_src or "",
            "frameSrcValue": self.frame_src or "",
            "mediaSrcValue": self.media_src or "",
            "objectSrcValue": self.object_src or "",
            "scriptSrcValue": self.script_src or "",
            "styleSrcValue": self.style_src or "",
            "baseUriValue": self.base_uri or "",
            "formActionValue": self.form_action or "",
            "frameAncestorsValue": self.frame_ancestors or "",
            "defaultSrcSecure": self.default_src_secure,
            "childSrcSecure": self.child_src_secure,
            "connectSrcSecure": self.connect_src_secure,
            "frameSrcSecure": self.frame_src_secure,
            "mediaSrcSecure": self.media_src_secure,
            "objectSrcSecure": self.object_src_secure,
            "scriptSrcSecure": self.script_src_secure,
            "styleSrcSecure": self.style_src_secure,
            "baseUriSecure": self.base_uri_secure,
            "formActionSecure": self.form_action_secure,
            "frameAncestorsSecure": self.frame_ancestors_secure,
        }



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from functools import cached_property
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagators import BasePropagator
from contrast.agent.assess.utils import (
    copy_events,
    get_last_event_from_sources,
    get_properties,
    track_string,
)
from contrast.utils.assess import tracking_util
from contrast.utils.assess.tag_utils import merge_tags
from contrast.utils.assess.formatting import FormatToken, StringToken
from contrast.utils.assess.formatting.tokenize_format import tokenize_format
from contrast.utils.assess.formatting.tokenize_cformat import (
    tokenize_format as tokenize_cformat,
)


class FormatPropagator(BasePropagator):
    def __init__(self, node, preshift, target):
        super().__init__(node, preshift, target)
        if node.method_name == "mod":
            self.tokens = tokenize_cformat(preshift.obj)
        else:
            self.tokens = tokenize_format(preshift.obj)
        self._tracked_args = None

    @cached_property
    def sources(self):
        # Any string that corresponds to a format token is a source
        sources = [
            token.get_arg(self.preshift.args, self.preshift.kwargs)
            for token in self.tokens
            if isinstance(token, FormatToken)
        ]
        return sources + [self.preshift.obj]

    @cached_property
    def tracked_sources(self):
        return any(tracking_util.recursive_is_tracked(src) for src in self.sources)

    def get_parents(self, *args):
        return get_last_event_from_sources(self.sources)

    @property
    def inputs_require_propagation(self):
        if tracking_util.recursive_is_tracked(self.preshift.obj):
            return True

        return self.tracked_sources

    def track_target(self):
        if (
            tracking_util.recursive_is_tracked(self.preshift.obj)
            or self.tracked_sources
        ):
            track_string(self.target)

    def _propagate(self):
        args, kwargs = self.preshift.args, self.preshift.kwargs

        # Offset in the target string
        target_offset = 0

        source_properties = get_properties(self.preshift.obj)

        for token in self.tokens:
            if isinstance(token, StringToken):
                self._propagate_formatted_string_properties(
                    token, target_offset, source_properties, self.target_properties
                )
                target_offset += len(token)
            else:  # FormatToken
                # Format the original input string, accounting for conversion
                # and spec, so that we know exactly how long it is in the
                # result string. We are in scope so this is safe
                orig_value, formatted = token.format(args, kwargs)
                self._propagate_formatted_string_properties(
                    token,
                    target_offset,
                    source_properties,
                    self.target_properties,
                    formatted=formatted,
                )
                self._propagate_format_specifier_properties(
                    orig_value, self.target_properties, target_offset
                )
                target_offset += len(formatted)

        if source_properties is not None:
            copy_events(self.target_properties, source_properties)

        merge_tags(self.target_properties.tags)

    def _propagate_formatted_string_properties(
        self, token, target_offset, source_properties, target_properties, formatted=None
    ):
        """
        Any tag that spans a format specifier will be expanded to
        cover the entirety of the formatted portion of the result
        """
        if source_properties:
            source_tags = source_properties.tags_at_range(token.span)

            token_len = len(formatted) if formatted is not None else len(token)

            for name in source_tags:
                end_offset = target_offset + token_len
                new_span = AdjustedSpan(target_offset, end_offset)
                target_properties.add_tag(name, new_span)

    def _propagate_format_specifier_properties(
        self, orig_value, target_properties, target_offset
    ):
        format_properties = get_properties(orig_value)
        if format_properties:
            for name, tag_ranges in format_properties.tags.items():
                for tag in tag_ranges:
                    # The span of the tag should not include any of the
                    # padding that is added by the formatter since it
                    # does not belong to the original string
                    new_span = tag.to_span(shift=target_offset)
                    target_properties.add_tag(name, new_span)

            copy_events(target_properties, format_properties)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.utils.string_utils import ensure_string

COMMON_SHELLS = [
    "sh",
    "csh",
    "tcsh",
    "scsh",
    "ksh",
    "dash",
    "bash",
    "zsh",
    "fish",
]


class SubprocessAction(DefaultAction):
    """
    Some functions in the subprocess module behave differently (from a security
    perspective) depending on several factors:
    - the value of the `shell` argument
    - whether arguments are passed as a string or as a list
    - if the command being executed is particularly dangerous (ie invokes a shell)

    If `shell` is explicitly set to True, all arguments are potentially vulnerable,
    whether they appear as a string or list.

    If `shell` is False (the default) and arguments are passed as a string, the entire
    string is potentially vulnerable. If arguments are passed as a list, only the first
    element is potentially vulnerable - unless the first element is a common shell
    command, in which case all arguments are potentially vulnerable.
    """

    unwind_source = False

    def extract_source(self, source, args, kwargs):
        return (
            source[0]
            if kwargs.get("shell") is not True
            and isinstance(source, (list, tuple))
            and len(source) > 0
            and ensure_string(source[0]) not in COMMON_SHELLS
            else source
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast.utils.decorators import fail_quietly
from contrast.utils.object_utils import get_name
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
)

original_concat_fns = {}


@fail_quietly()
def _patch_concat_fn(owner):
    """
    This is essentially a custom repatch. Jinja2 has several reexports of a `concat`
    function, which is set to `"".join`. If any module containing this function is
    loaded before our string patches are applied, we miss the patch on this attribute.
    Normally, repatching solves this problem; however repatching here doesn't work for
    multiple reasons:
    - `"".join` is already bound. We don't currently have machinery for repatching bound
      functions.
    - Repatching only covers module-level attributes, but in some instances, this is a
      class attribute

    The simplest solution is to overwrite these references with `"".join` again, after
    string patches have been applied.

    There are many reexports / aliases for `concat`, and these change across different
    versions of Jinja2. We currently only have fixes for those that appear to be
    critical for dataflow, but it's not impossible that we'll have to add more in the
    future.
    """
    if (original_concat := getattr(owner, "concat", None)) is None:
        return

    assert hasattr("".join, "__wrapped__"), (
        "string patches must be applied before jinja concat patches"
    )
    owner_name = get_name(owner)
    assert owner_name not in original_concat_fns
    if not (original_concat.__self__ == "" and original_concat.__name__ == "join"):
        raise RuntimeError(
            f"Unexpected value for {owner_name}.concat - skipping patch ({original_concat})"
        )

    owner.concat = "".join
    original_concat_fns[owner_name] = original_concat


def _reverse_patch_concat_fn(owner):
    if not hasattr(owner, "concat"):
        return

    owner_name = get_name(owner)
    original_concat = original_concat_fns.pop(owner_name, None)
    assert original_concat is not None
    owner.concat = original_concat


def patch_jinja2_environment(jinja2_environment_module):
    _patch_concat_fn(jinja2_environment_module)
    _patch_concat_fn(jinja2_environment_module.Environment)


def patch_jinja2_utils(jinja2_utils_module):
    _patch_concat_fn(jinja2_utils_module)


JINJA2_ENVIRONMENT_MODULE_NAME = "jinja2.environment"
JINJA2_UTILS_MODULE_NAME = "jinja2.utils"


def register_patches():
    register_module_patcher(patch_jinja2_environment, JINJA2_ENVIRONMENT_MODULE_NAME)
    register_module_patcher(patch_jinja2_utils, JINJA2_UTILS_MODULE_NAME)


def reverse_patches():
    unregister_module_patcher(JINJA2_ENVIRONMENT_MODULE_NAME)
    unregister_module_patcher(JINJA2_UTILS_MODULE_NAME)
    if (
        jinja2_environment_module := sys.modules.get(JINJA2_ENVIRONMENT_MODULE_NAME)
    ) is not None:
        _reverse_patch_concat_fn(jinja2_environment_module)
        _reverse_patch_concat_fn(jinja2_environment_module.Environment)
    if (jinja2_utils_module := sys.modules.get(JINJA2_UTILS_MODULE_NAME)) is not None:
        _reverse_patch_concat_fn(jinja2_utils_module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Any

from collections import OrderedDict
import sys
import time

from contrast.agent import scope
from contrast.utils.decorators import fail_quietly
from contrast.utils.string_utils import truncated_signature

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class _NoDefault:
    # unique value used for `pop()` override
    pass


class StringTracker(OrderedDict):
    AGE_OFF_THRESHOLD_SECS = 30

    def track(self, value):
        if not isinstance(value, (str, bytes, bytearray)):
            return None

        # ignore probable interned strings
        # TODO: PYT-2365 use is_interned here
        if len(value) < 2:
            return None

        if value and value not in self:
            from contrast.agent.assess.properties import Properties

            props = Properties(value)
            self.log_tracked(value)
            self[value] = props

        return self[value]

    def log_tracked(self, value):
        self._truncate_and_log("tracking new string", value)

    def log_ageoff(self, value):
        self._truncate_and_log("aging off string from tracker", value)

    def _truncate_and_log(self, msg, value):
        with scope.contrast_scope():
            logger.debug("%s: %s", msg, truncated_signature(value))

    def update_properties(self, value, properties):
        self[value] = properties

    def __delitem__(self, key):
        return super().__delitem__(id(key))

    def __getitem__(self, key):
        return super().__getitem__(id(key))

    def __setitem__(self, key, value):
        if key in self:
            return super(OrderedDict, self).__setitem__(id(key), value)
        return super().__setitem__(id(key), value)

    def get(self, key, default=None):
        return super().get(id(key), default)

    if sys.version_info[:2] >= (3, 11):
        # Prior to 3.11, `pop` likely uses `__getitem__` under the hood. However, this
        # is no longer true in 3.11, which means we need to override this method.
        def pop(self, key, default: Any = _NoDefault):
            if default is _NoDefault:
                return super().pop(id(key))
            return super().pop(id(key), default)

    def __contains__(self, key):
        return super().__contains__(id(key))

    def get_by_id(self, key_id):
        return super().__getitem__(key_id)

    @fail_quietly("Failure in string tracker ageoff")
    @scope.contrast_scope()
    def ageoff(self):
        """
        Age-off string tracker entries older than predefined threshold.

        NOTE: It is only approximately true that this OrderedDict's values are sorted by
        timestamp. They are in insertion-order, but timestamps can change due to
        circumstances that are currently not well-understood. Do not assume timestamps
        are in order.

        TODO: PYT-2365 - we believe incorrect string tracker timestamps may be affected
        by string interning. When this ticket is completed, check to make sure ageoff
        is still working as expected.
        """
        now = time.time()
        logger.debug("String tracker length (pre-ageoff): %s", len(self))
        # Keys are IDs in this loop
        for key in list(self.keys()):
            props = super().get(key)
            if (
                props is not None
                and (now - props.timestamp) > self.AGE_OFF_THRESHOLD_SECS
            ):
                self.log_ageoff(props.origin)
                try:  # noqa: SIM105 try-except is faster than suppress
                    super().__delitem__(key)
                except KeyError:
                    # Handles potential race condition that may occur if
                    # another thread may have already deleted the key after
                    # this thread gets the key but before it deletes it.
                    # More likely to happen in multi-threaded apps and if
                    # and app receives many requests per second.
                    pass



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.agent.middlewares.route_coverage.common import (
    build_signature,
    get_normalized_uri,
)

from aiohttp.web_urldispatcher import DynamicResource
from contrast_fireball import DiscoveredRoute


def create_aiohttp_routes(app) -> set[DiscoveredRoute]:
    """
    Returns all the routes registered to a AioHttp app as a dict
    :param app: AioHttp app instance
    :return: dict {route_id:  api.Route}
    """
    routes = set()

    for app_route in app.router._resources:
        for resource in app_route._routes:
            signature = build_signature(resource.handler.__name__, resource.handler)
            _route_attr = (
                app_route._formatter
                if isinstance(app_route, DynamicResource)
                else app_route._path
            )
            routes.add(
                DiscoveredRoute(
                    verb=resource.method,
                    url=get_normalized_uri(_route_attr),
                    signature=signature,
                    framework="AIOHTTP",
                )
            )

    return routes



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .django_httponly_rule import DjangoHttpOnlyRule
from .django_session_age_rule import DjangoSessionAgeRule
from .django_secure_flag_rule import DjangoSecureFlagRule
from .flask_httponly_rule import FlaskHttpOnlyRule
from .flask_session_age_rule import FlaskSessionAgeRule
from .flask_secure_flag_rule import FlaskSecureFlagRule

__all__ = [
    "DjangoHttpOnlyRule",
    "DjangoSessionAgeRule",
    "DjangoSecureFlagRule",
    "FlaskHttpOnlyRule",
    "FlaskSessionAgeRule",
    "FlaskSecureFlagRule",
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from functools import cached_property, lru_cache
from contrast.agent.framework import UNKNOWN_FRAMEWORK
from contrast.agent.request_context import RequestContext
from contrast_vendor.webob import Request

import contrast
from contrast.agent import agent_state, scope, request_state
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import (
    BaseMiddleware,
)
from contrast.agent.middlewares.environ_tracker import track_environ_sources
from contrast.utils.safe_import import safe_import_list
from contrast.utils.assess.duck_utils import safe_getattr_list

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")
DEFAULT_WSGI_NAME = "wsgi_app"


@lru_cache(maxsize=1)
def _get_flask_types():
    return tuple(safe_import_list("flask.Flask"))


class WSGIMiddleware(BaseMiddleware):
    """
    Contrast middleware; PEP-333(3) WSGI-compliant
    """

    @scope.contrast_scope()
    def __init__(
        self,
        wsgi_app,
        app_name=None,
        original_app=None,
        orig_pyramid_registry=None,
        *,
        framework_name: str = UNKNOWN_FRAMEWORK.name,
    ):
        # We need to keep the `original_app` `orig_pyramid_registry` kwarg for now to prevent a breaking API
        # change
        del original_app
        del orig_pyramid_registry

        # TODO: PYT-2852 Revisit application name detection
        self.app_name = (
            app_name
            if app_name is not None
            else safe_getattr_list(
                wsgi_app,
                [
                    "__name__",
                    "name",
                ],
                DEFAULT_WSGI_NAME,
            )
        )
        self.framework_name = framework_name
        agent_state.set_detected_framework(self.framework_name)

        super().__init__()

        if not agent_state.in_automatic_middleware() and isinstance(
            wsgi_app, _get_flask_types()
        ):
            # we need this to prevent a breaking API change when wrapping Flask apps
            wsgi_app = wsgi_app.wsgi_app

        self.wsgi_app = wsgi_app

    def __call__(self, environ, start_response):
        if request_state.get_request_id() is not None:
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request_id - passing through")
            return self.wsgi_app(environ, start_response)

        request_path = environ.get("PATH_INFO", "")
        request_method = environ.get("REQUEST_METHOD", "")

        with self.request_context_stack(
            request_method, request_path
        ) as request_context_stack:
            context = self.should_analyze_request(environ)
            if context:
                self.enter_agent_call_contexts(request_context_stack, context)
                return self.call_with_agent(context, environ, start_response)

            return self.call_without_agent(environ, start_response)

    @scope.contrast_scope()
    def call_with_agent(self, context: RequestContext, environ, start_response):
        track_environ_sources("wsgi", context, environ)

        try:
            self.prefilter(context)

            webob_request = Request(environ)
            with scope.pop_contrast_scope():
                try:
                    response = webob_request.get_response(self.wsgi_app)
                    context.extract_response(response)
                except Exception as exc:
                    context.response_exception = exc
                    raise

            self.postfilter(context)
            self.check_for_blocked(context)

            return response(environ, start_response)

        finally:
            self.handle_ensure(context, context.request)
            if context.assess_enabled:
                contrast.STRING_TRACKER.ageoff()

    @scope.contrast_scope()
    def call_without_agent(self, environ, start_response):
        """
        Normal without middleware call
        """
        return self.wsgi_app(environ, start_response)

    @cached_property
    def trigger_node(self):
        """
        WSGI-specific trigger node used by reflected xss postfilter rule

        The rule itself is implemented in the base middleware but we need to
        provide a WSGI-specific trigger node for reporting purposes.
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.wsgi_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "wsgi"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes


register_propagation_nodes(
    [
        {
            "module": "glob",
            "method_name": "translate",
            "source": "ARG_0,KWARG:pathname",
            "target": "RETURN",
            "action": "SPLAT",
        },
    ]
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections import defaultdict
from collections.abc import Mapping
from functools import cached_property
import re
from threading import RLock
from typing import Any, Callable

from contrast.agent.inventory.fingerprint import artifact_fingerprint
from contrast.utils.configuration_utils import (
    DEFAULT_PATHS,
    load_yaml_config,
)
from contrast_vendor import structlog as logging
from . import ConfigOption

from .agent import Agent
from .api import Api
from .application import Application
from .assess import Assess
from .inventory import Inventory
from .observe import Observe
from .protect import Protect
from .reporting import Reporting
from .root import Root
from .server import Server

logger = logging.getLogger("contrast")

DEFAULT_TEAMSERVER_TYPE = "EOP"
PUBLIC_SASS_TEAMSERVER_TYPES = {
    "app.contrastsecurity.com": "SAAS_DEFAULT",
    "app-agents.contrastsecurity.com": "SAAS_DEFAULT",
    "eval.contrastsecurity.com": "SAAS_POV",
    "eval-agents.contrastsecurity.com": "SAAS_POV",
    "ce.contrastsecurity.com": "SAAS_CE",
    "ce-agents.contrastsecurity.com": "SAAS_CE",
    "app.contrastsecurity.jp": "SAAS_TOKYO",
    "app-agents.contrastsecurity.jp": "SAAS_TOKYO",
    "cs": "SAAS_",
}

PRIVATE_SASS_TEAMSERVER_TYPE = "SAAS_CUSTOM"
PRIVATE_SASS_DOMAIN = ["contrastsecurity.com", "contrastsecurity.jp"]

TESTING_TEAMSERVER_TYPE = "TESTING"
TESTING_TEAMSERVER_NAMES = [
    "teamserver-darpa",
    "teamserver-darpa-agents",
    "alpha.contrastsecurity.com",
    "alpha-agents.contrastsecurity.com",
    "apptwo.contrastsecurity.com",
    "apptwo-agents.contrastsecurity.com",
    "teamserver-staging.contsec.jp",  # agents host name not defined yet.
    "teamserver-staging.contsec.com",
    "teamserver-staging-agents.contsec.com",
    "security-research.contrastsecurity.com",
    "security-research-agents.contrastsecurity.com",
    "teamserver-ops.contsec.com",  # agents host name not defined yet.
]

ConfigWatcher = Callable[["Mapping"], Any]


class AgentConfig(Mapping):
    def __init__(self):
        self.config_status = None
        self._config = {}
        self.watch_map: defaultdict[str, set[Watcher]] = defaultdict(set)
        self.updates_lock = RLock()
        self.build_configs()
        self.check_for_api_config()

    # Methods like `get`, `keys`, etc are automatically implemented by the Mapping abc.
    # They are derived from the concrete implementations below. See:
    # https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes
    # --- Required Mapping methods ---

    def __getitem__(self, key):
        return self._config[key].value()

    def __iter__(self):
        return iter(self._config)

    def __len__(self):
        return len(self._config)

    # --- End Required Mapping methods ---

    def get_option(self, key: str) -> ConfigOption | None:
        return self._config.get(key, None)

    def get_loggable_value(self, key: str) -> str | None:
        option = self.get_option(key)
        return option.loggable_value() if option else None

    def add_watcher(self, f: ConfigWatcher):
        """
        Calls f with the current config, and again whenever the config changes.
        As an optimization, f is only called again when the config value it uses changes.
        For example, if the config value of "agent.logger.level" changes, but f only uses
        "protect.enable", f will not be called again.
        """
        with self.updates_lock:
            Watcher(f)(self)

    def notify_watchers(self, keys: set[str]):
        """
        Notifies watchers registered for the given keys by reloading them.
        """
        with self.updates_lock:
            key_watchers = {f for key in keys for f in self.watch_map[key]}
            for f in key_watchers:
                f(self)

    def get_loggable_config(self):
        """
        The purpose of this function is to return a copy of the effective config with their resolved values rather than
        the ConfigOption object.
        """

        loggable_config = {}
        for key in self._config:
            loggable_config[key] = self.get_loggable_value(key)

        return loggable_config

    def log_config(self):
        loggable_config = self.get_loggable_config()

        logger.info("Current Configuration", **loggable_config)

    def build_configs(self):
        """
        For each class representing different parts of the config (Agent, etc),
        take the current config and apply any overriding logic such as overriding with
        environment keys/values.

        At this time the order of precedence is:
            os.env > Config yaml

        Meaning that a config value defined in os.environ must be used instead of
        the same config defined in contrast_security.yaml

        Note that CLI args (sys.argv) are not supported. This may change if the
        agent becomes a runner.
        """
        yaml_config = load_yaml_config(DEFAULT_PATHS)
        if not yaml_config:
            logger.info("No YAML config found; using default settings")

        builders = [
            Api,
            Agent,
            Application,
            Assess,
            Inventory,
            Observe,
            Protect,
            Reporting,
            Root,
            Server,
        ]
        for builder in builders:
            status = builder().build(self._config, yaml_config)
            if status:
                # An error here indicates a failure with either the type cast of the value
                # or an error on converting api.token to the 4 API config values
                self.config_status = status

    def check_for_api_config(self):
        """
        Validate api configurations were set.

        If any api configs are missing, log at ERROR and disable the agent.

        Returns the missing values or an empty list if none were missing.
        """

        api_keys = [
            "api.url",
            "api.service_key",
            "api.api_key",
            "api.user_name",
        ]

        missing_values = []
        for key in api_keys:
            option = self.get_option(key)
            if not option or option.value() in (None, ""):
                missing_values.append(key)

        if missing_values:
            msg = (
                f"Missing a required connection value for: {', '.join(missing_values)}"
            )
            logger.error(msg)
            self.config_status = "Unable to connect to Contrast, insufficient connection properties provided."
            enable_option = self.get_option("enable")
            enable_option.override_value = False

        return missing_values

    @property
    def session_id(self):
        return self.get("application.session_id")

    @session_id.setter
    def session_id(self, session_id):
        session_id_option = self.get_option("application.session_id")
        session_id_option.ui_value = session_id

        logger.debug(
            "Set session_id",
            session_id=session_id_option.value(),
            direct_to_teamserver=1,
        )

    def get_session_metadata(self):
        user_metadata = self.get("application.session_metadata", "")
        fingerprint = artifact_fingerprint()
        if not fingerprint:
            return user_metadata
        artifact_metadata = f"artifactHash={fingerprint}"
        return (
            artifact_metadata + "," + user_metadata
            if user_metadata
            else artifact_metadata
        )

    @property
    def app_code(self):
        return self.get("application.code")

    @property
    def app_metadata(self):
        return self.get("application.metadata")

    @property
    def app_group(self):
        return self.get("application.group")

    @property
    def app_tags(self):
        return self.get("application.tags")

    @property
    def assess_tags(self):
        return self.get("assess.tags")

    @cached_property
    def is_request_audit_enabled(self):
        return self.get("api.request_audit.enable")

    @cached_property
    def assess_enabled(self):
        return self.get("assess.enable")

    @cached_property
    def enable_automatic_middleware(self) -> bool:
        return self.get("agent.python.enable_automatic_middleware")

    @cached_property
    def application_settings_poll_interval(self):
        return self.get("agent.polling.app_settings_ms")

    @cached_property
    def server_settings_poll_interval(self):
        return self.get("agent.polling.server_settings_ms")

    @cached_property
    def heartbeat_poll_interval(self):
        return self.get("agent.polling.heartbeat_ms")

    @cached_property
    def teamserver_type(self):
        url = self.get("api.url")

        if not url:
            return ""

        for name, ts_type in PUBLIC_SASS_TEAMSERVER_TYPES.items():
            if name in url:
                # This regex matches https://cs001.contrastsecurity.com
                r_url = re.match(r".*(cs\d{3})\..*", url)
                if r_url:
                    # and gets the second group which is the matched in brackets cs001
                    # adds it to the SAAS_ type and becomes SAAS_CS001
                    numbers = r_url.group(1).upper()

                    return ts_type + numbers

                return ts_type

        for name in TESTING_TEAMSERVER_NAMES:
            if name in url:
                return TESTING_TEAMSERVER_TYPE

        for name in PRIVATE_SASS_DOMAIN:
            if name in url:
                return PRIVATE_SASS_TEAMSERVER_TYPE

        return DEFAULT_TEAMSERVER_TYPE

    def set_ui_value(self, key: str, value: object):
        """
        We have to be safe here with unknown configuration options as TeamServer can send us settings for which we do
        not have configurations. This most often happens when we're given configurations for rules that do not pertain
        to this agent.
        """
        if config_option := self.get_option(key):
            config_option.ui_value = value

    def update_ui_values(self, new_config: dict[str, object]) -> set[str]:
        """
        Update the UI values of the configuration options with the values from the new configuration.
        Watchers of the configuration options will be notified if the value of the configuration option changes.

        Returns the set of option names that changed.
        """
        with self.updates_lock:
            changed_keys = set()
            for key, value in new_config.items():
                # We have to be safe here with unknown configuration options as TeamServer can send us settings for which we do
                # not have configurations. This most often happens when we're given configurations for rules that do not pertain
                # to this agent.
                if config_option := self.get_option(key):
                    before = config_option.value()
                    config_option.ui_value = value
                    if before != config_option.value():
                        changed_keys.add(key)
            self.notify_watchers(changed_keys)
            return changed_keys


class Watcher:
    """
    A watcher is a function that is called when the configuration changes.

    The watcher wraps a function and spies on the configuration options that the function uses.
    It then registers itself in the configuration watch_map to be called when any of the
    configuration options change.
    """

    def __init__(self, f: ConfigWatcher):
        self.f = f

    def __call__(self, config: AgentConfig):
        config_spy = ConfigSpy(config)
        self.f(config_spy)
        if not config_spy._used_keys:
            logger.debug(
                "WARNING: Watcher does not use any config values. It will not be called when the config changes.",
                f=self.f,
            )

        for key in config_spy._used_keys:
            config.watch_map[key].add(self)


class ConfigSpy(Mapping):
    def __init__(self, wrapped: Mapping):
        self._wrapped = wrapped
        self._used_keys = set()

    def __getitem__(self, key):
        self._used_keys.add(key)
        return self._wrapped[key]

    def __iter__(self):
        return iter(self._wrapped)

    def __len__(self):
        return len(self._wrapped)

    def get_option(self, key: str) -> ConfigOption | None:
        if get_option := getattr(self._wrapped, "get_option", None):
            return get_option(key)
        return None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast.utils.configuration_utils import str_to_bool


class Reporting(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            # This is an internal setting to stop Fireball from connecting
            # to TeamServer.
            ConfigOption(
                canonical_name="reporting.contrast.enable",
                default_value=True,
                type_cast=str_to_bool,
            ),
            ConfigOption(
                canonical_name="reporting.logging.level",
                default_value=None,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="reporting.logging.stdout",
                default_value=False,
                type_cast=str_to_bool,
            ),
        ]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.agent.assess.tag import Tag
from contrast.utils.assess.tag_utils import intersection

from urllib.request import Request


class SsrfAction(DefaultAction):
    """
    Custom trigger action that implements SSRF rule logic.

    For SSRF, only tag ranges within specific regions of the URL are considered
    to be vulnerable. For a more detailed description, see the dataflow rule
    for SSRF in https://github.com/Contrast-Security-Inc/assess-specifications.
    """

    # Used to parse URLs into components for SSRF post-trigger analysis.
    # This follows the specification, linked above.
    SSRF_REGEX = (
        r"(?P<protocol>http|https|ftp|sftp|telnet|gopher|rtsp|rtsps|ssh|svn):\/\/"
        r"(?P<host>[^\/?]+)(?P<path>\/?[^?]*)(?P<querystring>\?.*)?"
    )

    # Protocol/host (NOT path/querystring) are the labs-approved URL components that trigger SSRF
    VULNERABLE_GROUPS = ["protocol", "host"]

    def find_all_tag_ranges(self, properties, desired_tags):
        vulnerable_tag_ranges = super().find_all_tag_ranges(properties, desired_tags)
        return self.trim_tag_ranges(properties.origin, vulnerable_tag_ranges)

    def find_any_tag_ranges(self, properties, desired_tags):
        vulnerable_tag_ranges = super().find_any_tag_ranges(properties, desired_tags)
        return self.trim_tag_ranges(properties.origin, vulnerable_tag_ranges)

    def trim_tag_ranges(self, source, tag_ranges):
        """
        Reduce the span of all vulnerable tags to only cover regions of the source URL
        that are actually vulnerable to SSRF.
        """
        if not isinstance(source, str):
            return tag_ranges

        match = re.match(self.SSRF_REGEX, source, flags=re.IGNORECASE)

        if match is None:
            return []

        ssrf_tag_ranges = []
        for group in self.VULNERABLE_GROUPS:
            start, end = match.span(group)
            if start != end:
                ssrf_tag_ranges.append(Tag(end - start, start))

        return intersection(tag_ranges, ssrf_tag_ranges)

    def extract_source(self, source, args, kwargs):
        """
        In py37, urllib.request does the same, via full_url()
        """
        if isinstance(source, Request):
            return source.full_url
        return source



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

from contrast_vendor import structlog as logging
from contrast.utils.locale import DEFAULT_ENCODING

logger = logging.getLogger("contrast")


def truncate(value, default="", length=256):
    """
    Truncate to `length` characters
    """
    if value is None:
        return default

    return value[:length]


def truncated_signature(value):
    """
    Get a log-friendly representation of a potentially long string. This function
    will truncate the string if necessary.

    First, we truncate the input to 60 characters - if this does happen, we'll also add
    `[TRUNCATED]` to the output. We then append the string's `id`. The string is
    converted to its most readable form using __repr__, which means that any newlines
    or invisible chars (like BEL) will be converted to something nice and readable.

    examples:
    'This is some string' (id=4443462824)
    'Here is a very long string that is longer th' [TRUNCATED] (id=4443294816)

    :param value: string whose truncated signature we want
    :return: string representation of the input value, truncated to 60 chars with
        its `id` appended as well. On any failure, return only the id.
    """
    try:
        append_truncate = ""

        if isinstance(value, bytearray):
            value = bytes(value)

        if len(value) > 60:
            value = value[:60]
            append_truncate = " [TRUNCATED]"

        value = ensure_string(value, errors="ignore")
        return f"{value!r}{append_truncate} (id={id(value)})"
    except Exception as e:
        logger.debug("Failed to truncate string: %s", e)
        return f"(id={id(value)})"


def index_of_any(value, search_chars):
    """
    Find the first index of a char in a string
    :param value: string to search
    :param search_chars: strings to search for
    :return: index if found else -1
    """

    for sc in search_chars:
        index = value.find(sc)

        if index != -1:
            return index

    return -1


def ends_with_any(value, strings):
    """
    Returns True if any of the strings are at the end of the value
    """
    return any(value.endswith(item) for item in strings)


def equals_ignore_case(this: str, that: str) -> bool:
    # PERF: this function is hot. Pre-checking str lengths is an optimization that
    # often prevents unnecessary (and slower) calls to lower().
    return len(this) == len(that) and this.lower() == that.lower()


def ensure_string(value: object, encoding=DEFAULT_ENCODING, errors="ignore") -> str:
    """
    Convert `value` of any type to a string, even if empty string.

    On failure given a custom encoding or using DEFAULT_ENCODING we
    return None and log a debug warning given the encoding that was used
    """
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        try:
            return value.decode(encoding, errors)
        except Exception:
            logger.debug(
                "WARNING: Failed to decode value using the encoding %s", encoding
            )
            return ""

    # Check to see if value can be made a str
    try:
        return str(value)
    except Exception:
        pass
    return ""


def ensure_binary(s: str | bytes, encoding="utf-8", errors="ignore") -> bytes:
    return s if isinstance(s, bytes) else s.encode(encoding, errors)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import platform
import time
from os import environ, path
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.install_lib import install_lib

UNSUPPORTED_PYTHON = (3, 15) if environ.get("CONTRAST_ALLOW_PY314") else (3, 14)

root_dir = path.abspath(path.dirname(__file__))


def read(*parts):
    with open(path.join(root_dir, *parts), encoding="utf-8") as f:
        return f.read()


def is_arm():
    machine = platform.machine()
    return "arm" in machine or "aarch" in machine


version_specifier = sys.version_info[:2]
if not version_specifier < UNSUPPORTED_PYTHON:
    raise RuntimeError(
        "Fatal: Cannot install contrast-agent: Unsupported python version "
        f"({platform.python_version()})"
    )

extension_path = path.join("contrast", "assess_extensions")
extension_source_dir = path.join("src", extension_path)
common_dir = path.join(extension_source_dir, "common")

if sys.platform.startswith("darwin"):
    link_args = ["-rpath", "@loader_path"]
    platform_args = []
else:
    platform_args = ["-Wno-cast-function-type"]
    link_args = []

debug = environ.get("ASSESS_DEBUG")
debug_args = ["-g", "-O1"] if debug else []
macros = [("ASSESS_DEBUG", "1")] if debug else []
macros.append(("EXTENSION_BUILD_TIME", f'"{time.ctime()}"'))

strict_build_args = ["-Werror"] if environ.get("CONTRAST_STRICT_BUILD") else []

c_sources = [
    path.join(common_dir, name)
    for name in [
        "patches.c",
        "scope.c",
        "logging.c",
        "intern.c",
        "propagate.c",
        "repr.c",
        "repeat.c",
        "streams.c",
        "subscript.c",
        "cast.c",
        "cs_str.c",
    ]
]
libraries = []


extensions = [
    Extension(
        "contrast.assess_extensions.cs_str",
        c_sources,
        libraries=libraries,
        include_dirs=[
            extension_source_dir,
            path.join(extension_source_dir, "include"),
        ],
        library_dirs=[extension_source_dir],
        # Path relative to the .so itself (works for gnu-ld)
        runtime_library_dirs=["$ORIGIN"],
        extra_compile_args=[
            "-Wall",
            "-Wextra",
            "-Wno-unused-parameter",
            "-Wmissing-field-initializers",
        ]
        + strict_build_args
        + debug_args
        + platform_args,
        extra_link_args=link_args,
        define_macros=macros,
    )
]


class ContrastBuildExt(build_ext):
    def run(self):
        build_ext.run(self)


class ContrastInstallLib(install_lib):
    def run(self):
        install_lib.run(self)


setup(
    cmdclass=dict(build_ext=ContrastBuildExt, install_lib=ContrastInstallLib),
    ext_modules=extensions,
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

import contrast

from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.utils import copy_from, track_string

from contrast.utils.patch_utils import (
    build_and_apply_patch,
    pack_self,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)
from contrast.utils.assess.stream_utils import ContrastFileProxy

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

SAFESTRING_MODULE_NAME = "django.utils.safestring"
UPLOADHANDLER_MODULE_NAME = "django.core.files.uploadhandler"
MARK_SAFE_NAME = "mark_safe"


def _get_source(args, kwargs):
    if args:
        return args[0]
    if kwargs:
        return kwargs.get("s")
    return None


def build_mark_safe_patch(orig_func, _):
    def mark_safe_patch(wrapped, instance, args, kwargs):
        """
        This patch implements a more optimized "deadzoned" original mark_safe call.

        We are deadzoning calling the original mark_safe because it may call SafeText,
        a class that inherits from str and that ends up propagating excessively.

        Because mark_safe is called many times when Django renders a template,
        excessive propagation is incredibly costly. So instead, we do the same work as
        the KEEP propagator would, but with a deadzoned call to the original function.
        """
        # if we're already in scope, don't bother doing any analysis
        if scope.in_contrast_or_propagation_scope():
            return wrapped(*args, **kwargs)

        # if we're not yet in scope, call wrapped and analysis in scope
        with scope.contrast_scope():
            # We don't wrap wrapped call in try/catch because if it raises an error
            # we don't want to do any analysis.
            result = wrapped(*args, **kwargs)

            if contrast.REQUEST_CONTEXT.get() is not None:
                try:
                    logger.debug("Analyzing in %s custom propagator.", MARK_SAFE_NAME)
                    source = _get_source(pack_self(instance, args), kwargs)
                    track_string(result)
                    copy_from(result, source)
                except Exception as e:
                    logger.debug(
                        "Failed to analyse in %s propagator. %s", MARK_SAFE_NAME, str(e)
                    )

        return result

    return wrap_and_watermark(orig_func, mark_safe_patch)


def build_new_file_patch(orig_func, patch_policy):
    def new_file_patch(*args, **kwargs):
        """
        Patch for TemporaryFileUploadHandler.new_file

        Uploaded files that exceed the FILE_UPLOAD_MAX_MEMORY_SIZE threshold are
        streamed from a temporary file on disk rather than loaded into memory.

        This patch enables us to track the contents of uploaded files that exceed this
        threshold.

        - args[0] is self, which is an instance of TemporaryFileUploadHandler
        - args[0].file is a TemporaryUploadedFile, which is the stream we want to proxy
        """
        result = orig_func(*args, **kwargs)

        if args and hasattr(args[0], "file"):
            try:
                # This also marks the stream as a source While ContrastFileProxy was
                # originally intended for file objects in Py27, it also meets our needs
                # for proxying the TemporaryUploadedFile stream.  Eventually we may
                # decide to use a different proxy class, or at least a different name
                # (or change the docstring in ContrastFileProxy).
                args[0].file = ContrastFileProxy(args[0].file)
            except Exception as e:
                logger.debug(
                    "Failed to track large uploaded file as source stream: %s", str(e)
                )

        return result

    return new_file_patch


def patch_django_safestring(safestring_module):
    build_and_apply_patch(safestring_module, MARK_SAFE_NAME, build_mark_safe_patch)


def patch_django_uploadhandler(uploadhandler_module):
    build_and_apply_patch(
        uploadhandler_module.TemporaryFileUploadHandler,
        "new_file",
        build_new_file_patch,
    )


def register_patches():
    register_module_patcher(patch_django_safestring, SAFESTRING_MODULE_NAME)
    register_module_patcher(patch_django_uploadhandler, UPLOADHANDLER_MODULE_NAME)


def reverse_patches():
    unregister_module_patcher(SAFESTRING_MODULE_NAME)
    safestring_module = sys.modules.get(SAFESTRING_MODULE_NAME)
    if safestring_module:
        patch_manager.reverse_patches_by_owner(safestring_module)

    unregister_module_patcher(UPLOADHANDLER_MODULE_NAME)
    uploadhandler_module = sys.modules.get(UPLOADHANDLER_MODULE_NAME)
    if uploadhandler_module:
        patch_manager.reverse_patches_by_owner(
            uploadhandler_module.TemporaryFileUploadHandler
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum, auto

from contrast_agent_lib import constants


class DocumentType(Enum):
    NORMAL = auto()
    JSON = auto()
    XML = auto()


class InputType(Enum):
    COOKIE_NAME = auto()
    COOKIE_VALUE = auto()
    HEADER = auto()
    PARAMETER_NAME = auto()
    PARAMETER_VALUE = auto()
    URI = auto()
    JSON_VALUE = auto()
    MULTIPART_NAME = auto()
    XML_VALUE = auto()
    UNKNOWN = auto()
    METHOD = auto()
    URL_PARAMETER = auto()

    def cef_string(self, key: str):
        fmt = _CEF_FMT_FROM_INPUT_TYPE.get(self, "untrusted input")
        return fmt.format(key) if fmt.endswith("{}") else fmt

    @property
    def is_body_based(self) -> bool:
        """
        Indicates if the input can be found in the body of a request.

        This doesn't guarntee that the input is actually from the body,
        for example, a querystring parameter will return True because we
        don't distinguish between querystring and body form parameters.
        """
        return self in (
            InputType.PARAMETER_VALUE,
            InputType.MULTIPART_NAME,
            InputType.JSON_VALUE,
            InputType.XML_VALUE,
        )

    @property
    def agent_lib_type(self) -> int:
        return _AGENT_LIB_INPUT_TYPE[self]

    @classmethod
    def from_agent_lib_type(cls, type: int):
        return cls(_REVERSE_AGENT_LIB_INPUT_TYPE[type])


_AGENT_LIB_INPUT_TYPE = {
    InputType.COOKIE_NAME: constants.InputType["CookieName"],
    InputType.COOKIE_VALUE: constants.InputType["CookieValue"],
    InputType.HEADER: constants.InputType["HeaderValue"],
    InputType.PARAMETER_NAME: constants.InputType["ParameterKey"],
    InputType.PARAMETER_VALUE: constants.InputType["ParameterValue"],
    InputType.URI: constants.InputType["UriPath"],
    InputType.JSON_VALUE: constants.InputType["JsonValue"],
    InputType.MULTIPART_NAME: constants.InputType["MultipartName"],
    InputType.XML_VALUE: constants.InputType["XmlValue"],
    InputType.METHOD: constants.InputType["Method"],
    InputType.URL_PARAMETER: constants.InputType["UrlParameter"],
}
_REVERSE_AGENT_LIB_INPUT_TYPE = {v: k for k, v in _AGENT_LIB_INPUT_TYPE.items()} | {
    constants.InputType["JsonKey"]: InputType.JSON_VALUE
}


_CEF_FMT_FROM_INPUT_TYPE = {
    InputType.COOKIE_NAME: "cookie {}",
    InputType.COOKIE_VALUE: "cookie {}",
    InputType.HEADER: "header {}",
    InputType.PARAMETER_NAME: "parameter {}",
    InputType.PARAMETER_VALUE: "parameter {}",
    InputType.URI: "URI",
    InputType.JSON_VALUE: "JSON value {}",
    InputType.MULTIPART_NAME: "name of the multipart {}",
    InputType.XML_VALUE: "XML value {}",
    InputType.METHOD: "method {}",
    InputType.UNKNOWN: "untrusted input",
}


@dataclass
class UserInput:
    value: str
    type: InputType = InputType.UNKNOWN
    name: str | None = None
    path: str | None = None
    matcher_ids: list[str] = field(default_factory=list)
    document_type: DocumentType = DocumentType.NORMAL

    def __post_init__(self):
        if self.type == InputType.JSON_VALUE:
            self.document_type = DocumentType.JSON
            self.path = self.name
        elif self.type == InputType.XML_VALUE:
            self.document_type = DocumentType.XML
            self.path = self.name

    @property
    def is_name_based(self) -> bool:
        """
        Indicates if the input is based on a name (e.g., parameter name, cookie name).
        """
        return self.type in (
            InputType.COOKIE_NAME,
            InputType.PARAMETER_NAME,
            InputType.MULTIPART_NAME,
        )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


from typing import Any
import contrast

import xml.etree.ElementTree
from xml.etree.ElementTree import ParseError

from contrast.agent.exclusions import Exclusions
from contrast.agent.request_context import RequestContext
from contrast.api.user_input import DocumentType
from contrast.agent.settings import Settings
from contrast.agent.agent_lib import input_tracing
from contrast_vendor import structlog as logging
from contrast.utils.decorators import fail_quietly
from contrast_agent_lib import constants

logger = logging.getLogger("contrast")


def _get_enabled_rules() -> int:
    """
    This converts our list of enabled rules to an integer value as the bitmask that the
    Agent Library expects.
    """
    rules = 0
    settings = Settings()

    for rule_tuple in settings.protect_rules.items():
        if (
            rule_tuple
            and rule_tuple[1]
            and rule_tuple[1].enabled
            and rule_tuple[1].RULE_NAME in constants.RuleType
        ):
            rules |= constants.RuleType[rule_tuple[1].RULE_NAME]
    return rules


def analyze_inputs() -> None:
    """
    Perform input analysis through agent-lib. Results are stored on
    context.user_input_analysis, which is reset every time this function is called.

    Some rules have a special "worth watching" analysis mode. In prefilter, we use this
    more liberal mode to ensure we don't miss attacks that should be blocked at trigger
    time. However, if we make it to the end of the request (without raising a
    SecurityException), we redo input analysis with worth watching mode disabled, which
    leads to more accurate PROBED results (fewer PROBED FPs).
    """
    context = contrast.REQUEST_CONTEXT.get()
    if context is None:
        return

    context.user_input_analysis = []

    rules = _get_enabled_rules()

    # Input analysis for all the input_tracing.InputType enum values
    _evaluate_headers(context, rules)
    _evaluate_cookies(context, rules)
    _evaluate_body(context, rules)
    _call_check_method_tampering(context)
    _evaluate_query_string_params(context, rules)
    _call_agent_lib_evaluate_input(
        constants.InputType["UriPath"],
        context.request.path,
        rules,
        context,
    )
    _evaluate_path_params(context, rules)
    _evaluate_multipart_request(context, rules)


def _evaluate_headers(context: RequestContext, rules: int) -> None:
    for header_name, header_value in context.request.headers.items():
        if "cookie" in header_name.lower() or check_param_input_exclusions(
            context.exclusions, "HEADER", header_name
        ):
            continue

        input_analysis = input_tracing.evaluate_header_input(
            header_name,
            header_value,
            rules,
            prefer_worth_watching=True,
        )

        if input_analysis:
            context.user_input_analysis.extend(input_analysis)
            # Report and block attack if necessary
            _report_and_block_at_perimeter(
                input_analysis,
                ["bot-blocker", "reflected-xss", "unsafe-file-upload"],
                context,
            )


def _evaluate_cookies(context: RequestContext, rules: int) -> None:
    for cookie_name, cookie_value in context.request.cookies.items():
        if check_param_input_exclusions(context.exclusions, "COOKIE", cookie_name):
            continue

        _call_agent_lib_evaluate_input(
            constants.InputType["CookieName"],
            cookie_name,
            rules,
            context,
        )
        _call_agent_lib_evaluate_input(
            constants.InputType["CookieValue"],
            cookie_value,
            rules,
            context,
            input_key=cookie_name,
        )


@fail_quietly("Failed to evaluate body")
def _evaluate_body(context: RequestContext, rules: int) -> None:
    if not context.request.is_body_readable:
        return
    if check_url_input_exclusion(context.exclusions, "BODY", context.request.url):
        return

    body_type = context.request._get_document_type()
    if body_type == DocumentType.JSON:
        try:
            json_body = context.request.json
        except Exception as e:
            logger.debug("WARNING: Failed to parse JSON in request body", exc_info=e)
            return
        _evaluate_body_json(context, rules, json_body)
    elif body_type == DocumentType.XML:
        try:
            data = xml.etree.ElementTree.fromstring(context.request.body)
        except ParseError as e:
            logger.debug("WARNING: Failed to parse XML in request body", exc_info=e)
            return

        text_list = [element.text for element in data]

        for text in text_list:
            if not str(text).startswith("\n"):
                _call_agent_lib_evaluate_input(
                    constants.InputType["XmlValue"],
                    str(text),
                    rules,
                    context,
                )
    else:
        _evaluate_key_value_parameters(context, rules, querystring=False)


def _evaluate_body_json(context: RequestContext, rules: int, body: Any) -> None:
    # Using recursion for now to get all the json values and keys and pass them
    # through agent_lib until agent_lib implements parsing of the body for python
    if isinstance(body, dict):
        for key, value in body.items():
            _call_agent_lib_evaluate_input(
                constants.InputType["JsonKey"],
                key,
                rules,
                context,
            )
            _evaluate_body_json(context, rules, value)
    elif isinstance(body, list):
        for item in body:
            _evaluate_body_json(context, rules, item)
    elif isinstance(body, str):
        _call_agent_lib_evaluate_input(
            constants.InputType["JsonValue"],
            body,
            rules,
            context,
        )


def _evaluate_query_string_params(context: RequestContext, rules: int) -> None:
    """
    Get agent-lib input analysis for all query parameters. This information is stored on
    request context.
    """
    if check_url_input_exclusion(
        context.exclusions, "QUERYSTRING", context.request.url
    ):
        return

    _evaluate_key_value_parameters(context, rules, querystring=True)


def _evaluate_key_value_parameters(
    context: RequestContext, rules: int, *, querystring: bool
) -> None:
    """
    Used for both form parameters (from the request body) and query string parameters
    """
    param_dict = context.request.GET if querystring else context.request.POST

    for param_key, param_value in param_dict.items():
        if not isinstance(param_value, str):
            continue

        _call_agent_lib_evaluate_input(
            constants.InputType["ParameterKey"],
            param_key,
            rules,
            context,
            input_key=param_key,
        )
        _call_agent_lib_evaluate_input(
            constants.InputType["ParameterValue"],
            param_value,
            rules,
            context,
            input_key=param_key,
        )


def _evaluate_path_params(context: RequestContext, rules: int) -> None:
    """
    Get agent-lib input analysis for all path parameters. This information is
    stored on request context.
    """
    for param in context.request.get_url_parameters():
        if check_param_input_exclusions(context.exclusions, "PARAMETER", param):
            continue

        _call_agent_lib_evaluate_input(
            constants.InputType["UrlParameter"],
            param,
            rules,
            context,
        )


def _evaluate_multipart_request(context: RequestContext, rules: int):
    """
    This is refering to Content-Type: multipart/form-data and checking the file_name for every
    multipart request if there is none it checks the name
    """
    for key, value in context.request.get_multipart_headers().items():
        if value is None and key is None:
            continue

        multipart_name = value if value is not None else key
        _call_agent_lib_evaluate_input(
            constants.InputType["MultipartName"],
            multipart_name,
            rules,
            context,
        )


def _call_check_method_tampering(context: RequestContext) -> None:
    input_analysis_value = input_tracing.check_method_tampering(context.request.method)

    if input_analysis_value:
        context.user_input_analysis.extend(input_analysis_value)
        _report_and_block_at_perimeter(
            input_analysis_value, ["reflected-xss", "unsafe-file-upload"], context
        )


def _call_agent_lib_evaluate_input(
    input_type: int,
    input_value: str,
    rule_set: int,
    context: RequestContext,
    *,
    input_key="",
) -> None:
    input_analysis_results = input_tracing.evaluate_input_by_type(
        input_type, input_value, input_key, rule_set, prefer_worth_watching=True
    )

    if input_analysis_results:
        context.user_input_analysis.extend(input_analysis_results)
        _report_and_block_at_perimeter(
            input_analysis_results, ["reflected-xss", "unsafe-file-upload"], context
        )


def _report_and_block_at_perimeter(
    input_analysis: list[input_tracing.InputAnalysisResult],
    perimeter_rule_names: list[str],
    context: RequestContext,
) -> None:
    """
    Checks a list of rules and reports if it finds a score(int with value 0-100 indicating percentage
    of certainty of attack) higher than 90 and blocks if the agent is configured in block mode.
    """
    settings = Settings()
    for result in input_analysis:
        if result.rule_id not in perimeter_rule_names:
            continue
        if rule := settings.protect_rules.get(result.rule_id):
            if attack := rule.build_attack_with_match(result.input.value, result, None):
                context.attacks.append(attack)

            if result.score >= 90:
                logger.debug(
                    f"Input analysis found a value '{result.input.value}' "
                    f"that violated {rule.RULE_NAME} rule!"
                )

                if rule.is_blocked():
                    raise contrast.SecurityException(rule_name=rule.RULE_NAME)


def check_url_input_exclusion(
    exclusions: Exclusions | None, input_type: str, input_name: str
) -> bool:
    if not exclusions:
        return False

    return exclusions.evaluate_input_exclusions_url(
        exclusions, input_type, input_name, mode="defend"
    )


def check_param_input_exclusions(
    exclusions: Exclusions | None, input_type: str, input_name: str
) -> bool:
    if not exclusions:
        return False

    return exclusions.evaluate_input_exclusions(
        exclusions, input_type, input_name, mode="defend"
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule
from contrast.agent.policy.utils import CompositeNode


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_REDOS",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_REDOS",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
]


# This rule, like SSRF, has an accompanying action to enforce an extra level
# of assertion on the violation, so it may not always report, depending on the
# value reported.
# Like the propagation methods for Re::Match, it also requires custom patching
# to instrument properly.
# For all of these methods, ARG_0 is the pattern passed in to the re method,
# which we verify against the REDOS action for exploitability but we do not
# care about its origin so it doesn't matter if it's tracked.
redos_triggers = [
    CompositeNode(
        {
            "module": "re",
            "action": "REDOS",
        },
        [
            {
                "method_name": "match",
                # ARG_1 is the user input which we track and verify if untrusted
                "source": "ARG_1,KWARG:string",
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "match",
                # ARG_1 is the user input which we track and verify if untrusted
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_1,KWARG:string,KWARG:pattern",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "search",
                # ARG_1 is the user input which we track and verify if untrusted
                "source": "ARG_1,KWARG:string",
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "search",
                # ARG_1 is the user input which we track and verify if untrusted
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_1,KWARG:string,KWARG:pattern",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "finditer",
                # ARG_1 is the user input which we track and verify if untrusted
                "source": "ARG_1,KWARG:string",
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "finditer",
                # ARG_1 is the user input which we track and verify if untrusted
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_1,KWARG:string",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "findall",
                # ARG_1 is the user input which we track and verify if untrusted
                "source": "ARG_1,KWARG:string",
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "findall",
                # ARG_1 is the user input which we track and verify if untrusted
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_1,KWARG:string,KWARG:source",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "fullmatch",
                # ARG_1 is the user input which we track and verify if untrusted
                "source": "ARG_1,KWARG:string",
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "fullmatch",
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_1,KWARG:string",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "sub",
                # ARG_2 is the user input which we track and verify if untrusted
                "source": "ARG_2,KWARG:string",
                "policy_patch": False,
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "sub",
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_2,KWARG:string",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "subn",
                # ARG_2 is the user input which we track and verify if untrusted
                "source": "ARG_2,KWARG:string",
                "policy_patch": False,
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "subn",
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_2,KWARG:string",
                "policy_patch": False,
            },
            {
                "module": "re",
                "method_name": "split",
                "source": "ARG_1,KWARG:string,KWARG:source",
            },
            {
                "module": "re",
                "class_name": "Pattern",
                "method_name": "split",
                # NOTE: because the redos action needs ARG_0 to analyze the pattern, the fact
                # this is an instance method node is ignored in the trigger logic.
                "source": "ARG_1,KWARG:string",
                "policy_patch": False,
            },
        ],
    )
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "redos",
        redos_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from contrast.agent.settings import Settings
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.assess.rules import build_finding


class TriggerConfigRule(DataflowRule):
    """
    Base class for config rules that are actually implemented as triggers

    We have several rules that are technically config rules in TS, but are implemented
    as triggers for some of the frameworks we support. In these cases, the rule itself
    is fired like a trigger/dataflow rule. However, we need to structure the data we
    send to TS in a way that looks like a config rule.
    """

    SESSION_ID = "sessionId"
    PATH = "path"
    SNIPPET = "snippet"

    def _find_first_app_frame(self, node, stack):
        """
        Look for the first stack frame that doesn't belong to the framework

        This is a heuristic and may not work in all cases.
        """
        for frame in stack:
            if not frame.file.startswith(node.module):
                return frame

        # Fallback case
        return stack[0]

    def _create_filename(self, frame):
        filename = frame.file.rstrip(".py").replace(".", os.sep) + ".py"
        return f"{filename}:{frame.line_number}"

    def _create_snippet(self, node, target, event):
        # Rely only on kwarg param for reporting
        source = [x for x in node.sources if isinstance(x, str)][0]
        params = f"{source}={target}"

        method_name = (
            f"{event.stack[0].file.rstrip('.py')}.{event.stack[0].method}"
            if event.stack[0].file.startswith(node.module)
            else node.name
        )

        return f"{method_name}({params})"

    def create_finding(self, orig_properties, node, target, events, **kwargs):
        """
        Create a finding that makes the trigger rule look like a config rule in TS
        """
        settings = Settings()

        properties = {}
        properties[self.SESSION_ID] = settings.config.session_id

        reported_frame = self._find_first_app_frame(node, events[-1].stack)
        properties[self.PATH] = self._create_filename(reported_frame)
        properties[self.SNIPPET] = self._create_snippet(node, target, events[-1])

        return build_finding(self, properties, **kwargs)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


import ctypes
from dataclasses import dataclass, replace
import functools
from enum import IntEnum
from typing import Callable

from contrast.agent import agent_lib
from contrast.api.user_input import InputType, UserInput
from contrast_vendor import structlog as logging
from contrast_agent_lib import lib_contrast, constants

logger = logging.getLogger("contrast")

# These are rules we do not have an implementation for yet
# Other rule IDs where added directly into the protect rule class
SSJS_INJECTION_RULE_ID = 1 << 7

# This should not be used directly. It's used to construct INPUT_TYPE_LOOKUP below.
_AGENT_LIB_TO_TS_INPUT_TYPE = {
    "CookieName": InputType.COOKIE_NAME,
    "CookieValue": InputType.COOKIE_VALUE,
    "HeaderKey": InputType.HEADER,
    "HeaderValue": InputType.HEADER,
    "JsonKey": InputType.JSON_VALUE,  # there is no such thing as InputType.JSON_KEY
    "JsonValue": InputType.JSON_VALUE,
    "Method": InputType.METHOD,
    "ParameterKey": InputType.PARAMETER_NAME,
    "ParameterValue": InputType.PARAMETER_VALUE,
    "UriPath": InputType.URI,
    "UrlParameter": InputType.URL_PARAMETER,
    "MultipartName": InputType.MULTIPART_NAME,
    "XmlValue": InputType.XML_VALUE,
}

_RULE_ID_LOOKUP = {
    rule_int: rule_str for rule_str, rule_int in constants.RuleType.items()
}


class DBType(IntEnum):
    DB2 = 1
    MYSQL = 2
    ORACLE = 3
    POSTGRES = 4
    SQLITE = 5
    SQL_SERVER = 6
    UNKNOWN = 7

    @staticmethod
    def from_str(label: str):
        label = label.upper()
        try:
            return DBType[label]
        except KeyError:
            if label == "SQLITE3":
                return DBType.SQLITE
            if label == "POSTGRESQL":
                return DBType.POSTGRES
            if label in ("SQL SERVER", "SQL_SERVER", "SQLSERVER"):
                return DBType.SQL_SERVER

            return DBType.UNKNOWN


@functools.lru_cache(maxsize=1)
def input_type_lookup() -> dict[int, InputType]:
    """
    agent lib input type int -> InputType enum (TS representation)

    we'd like this to be a module-level dict, but we need to load this after agent_lib
    initialization
    """
    return {
        input_type_int: _AGENT_LIB_TO_TS_INPUT_TYPE.get(
            input_type_str, InputType.UNKNOWN
        )
        for input_type_str, input_type_int in constants.InputType.items()
    }


@dataclass
class InputAnalysisResult:
    input: UserInput
    rule_id: str
    score: float
    attack_count: int = 0

    @classmethod
    def from_ceval_result(cls, input: UserInput, ceval_result: constants.CEvalResult):
        return cls(
            input=replace(
                input,
                matcher_ids=InputAnalysisResult._extract_matcher_ids(ceval_result),
            ),
            rule_id=_RULE_ID_LOOKUP[ceval_result.rule_id],
            score=ceval_result.score,
        )

    @staticmethod
    def _extract_matcher_ids(ceval_result: constants.CEvalResult) -> list[str]:
        matched_ids_buf = bytes(ceval_result.matched_ids)
        return [
            match.decode()
            for match in matched_ids_buf[: matched_ids_buf.find(b"\0\0")].split(b"\0")
            if match
        ]

    def fully_evaluate(self) -> list[InputAnalysisResult]:
        """
        Evaluates the input using the same rule and input type, but without
        optimizations such as worth watching preferences. This is slower, but
        more accurate.
        """
        return (
            evaluate_header_input(
                self.input.name,
                self.input.value,
                constants.RuleType[self.rule_id],
                prefer_worth_watching=False,
            )
            if self.input.type == InputType.HEADER
            else evaluate_input_by_type(
                self.input.type.agent_lib_type,
                self.input.value,
                self.input.name,
                constants.RuleType[self.rule_id],
                prefer_worth_watching=False,
            )
        )


class InjectionResult:
    def __init__(
        self,
        user_input: str,
        input_index: int,
        input_len: int,
        ccheck_query_sink_result: agent_lib.CCheckQuerySinkResult,
    ):
        self.boundary_overrun_index = ccheck_query_sink_result.boundary_overrun_index
        self.end_index = ccheck_query_sink_result.end_index
        self.input_boundary_index = ccheck_query_sink_result.input_boundary_index
        self.start_index = ccheck_query_sink_result.start_index
        self.user_input = user_input
        self.input_index = input_index
        self.input_len = input_len


def evaluate_header_input(
    header_name: str, header_value: str, rules: int, prefer_worth_watching: bool
) -> list[InputAnalysisResult]:
    evaluations = []

    if not agent_lib.IS_INITIALIZED:
        return evaluations

    if rules == 0:
        return evaluations

    def is_valid_return(code):
        return code == 0

    name = ctypes.c_char_p(bytes(str(header_name), "utf8"))
    value = ctypes.c_char_p(bytes(str(header_value), "utf8"))
    results_len = ctypes.c_size_t()
    results = ctypes.POINTER(constants.CEvalResult)()

    ret = agent_lib.call(
        lib_contrast.evaluate_header_input,
        is_valid_return,
        name,
        value,
        rules,
        prefer_worth_watching,
        ctypes.byref(results_len),
        ctypes.byref(results),
    )

    map_result_and_free_eval_result(
        ret,
        results,
        results_len,
        InputType.HEADER,
        header_name,
        header_value,
        is_valid_return,
        evaluations,
    )
    return evaluations


def evaluate_input_by_type(
    input_type: int,
    input_value: str,
    input_key: str | None,
    rules: int,
    prefer_worth_watching: bool,
) -> list[InputAnalysisResult]:
    evaluations = []

    if not agent_lib.IS_INITIALIZED:
        return evaluations

    if rules == 0:
        return evaluations

    def is_valid_return(code):
        return code == 0

    if not isinstance(input_value, str):
        input = ctypes.c_char_p(bytes(str(input_value), "utf8"))
    else:
        input = ctypes.c_char_p(bytes(input_value, "utf8"))
    long_input_type = ctypes.c_long(input_type)
    results_len = ctypes.c_size_t()
    results = ctypes.POINTER(constants.CEvalResult)()

    ret = agent_lib.call(
        lib_contrast.evaluate_input,
        is_valid_return,
        input,
        long_input_type,
        rules,
        prefer_worth_watching,
        ctypes.byref(results_len),
        ctypes.byref(results),
    )

    map_result_and_free_eval_result(
        ret,
        results,
        results_len,
        InputType.from_agent_lib_type(input_type),
        input_key,
        input_value,
        is_valid_return,
        evaluations,
    )
    return evaluations


def check_method_tampering(method: str) -> list[InputAnalysisResult]:
    evaluations = []

    if not agent_lib.IS_INITIALIZED:
        return evaluations

    # These codes should match https://agent-lib.prod.dotnet.contsec.com/src/contrast_c/method_tampering.rs.html#11-28
    IS_TAMPERING = 1
    IS_NOT_TAMPERING = 0

    def is_valid_return(code):
        return code in (IS_TAMPERING, IS_NOT_TAMPERING)

    name = ctypes.c_char_p(bytes(method, "utf8"))

    ret = agent_lib.call(
        lib_contrast.is_method_tampering,
        is_valid_return,
        name,
    )

    if ret != IS_TAMPERING:
        return []

    c_eval_res = constants.CEvalResult()
    c_eval_res.rule_id = constants.RuleType.get("method-tampering")
    c_eval_res.input_type = constants.InputType.get("Method")
    c_eval_res.score = 100

    return [
        InputAnalysisResult.from_ceval_result(
            UserInput(
                type=InputType.METHOD,
                value=method,
                name=method,  # name is needed for CEF logging
            ),
            c_eval_res,
        )
    ]


def check_sql_injection_query(
    user_input_start_index: int,
    user_input_len: int,
    db_type: DBType,
    built_sql_query: str,
) -> InjectionResult | None:
    if not agent_lib.IS_INITIALIZED:
        return None

    def is_valid_return(code):
        return -1 <= code <= 0

    input_index = ctypes.c_uint32(user_input_start_index)
    input_len = ctypes.c_uint32(user_input_len)
    c_db_type = ctypes.c_uint32(int(db_type))
    sql_query = ctypes.c_char_p(bytes(built_sql_query, "utf8"))
    results = ctypes.pointer(agent_lib.CCheckQuerySinkResult())

    ret = agent_lib.call(
        lib_contrast.check_sql_injection_query,
        is_valid_return,
        input_index,
        input_len,
        c_db_type,
        sql_query,
        ctypes.byref(results),
    )

    evaluation = map_result_and_free_check_query_sink_result(
        ret,
        results,
        built_sql_query,
        user_input_start_index,
        user_input_len,
        is_valid_return,
    )
    return evaluation


def check_cmd_injection_query(
    user_input_start_index: int, user_input_len: int, user_input_txt: str
) -> InjectionResult | None:
    if not agent_lib.IS_INITIALIZED:
        return None

    def is_valid_return(code):
        return -1 <= code <= 0

    input_index = ctypes.c_uint32(user_input_start_index)
    input_len = ctypes.c_uint32(user_input_len)
    cmd_text = ctypes.c_char_p(bytes(user_input_txt, "utf8"))
    results = ctypes.POINTER(agent_lib.CCheckQuerySinkResult)()

    ret = agent_lib.call(
        lib_contrast.check_cmd_injection_query,
        is_valid_return,
        input_index,
        input_len,
        cmd_text,
        ctypes.byref(results),
    )

    evaluation = map_result_and_free_check_query_sink_result(
        ret,
        results,
        user_input_txt,
        user_input_start_index,
        user_input_len,
        is_valid_return,
    )
    return evaluation


def map_result_and_free_eval_result(
    ret: int,
    results: ctypes._Pointer[constants.CEvalResult],
    results_len: ctypes.c_size_t,
    type: InputType,
    key: str | None,
    value: str,
    is_valid_return: Callable[[int], bool],
    evaluations: list[InputAnalysisResult],
) -> None:
    if ret == 0 and bool(results) and results_len.value > 0:
        evaluations.extend(
            InputAnalysisResult.from_ceval_result(
                UserInput(type=type, name=key, value=value), results[i]
            )
            for i in range(results_len.value)
        )

        # ctypes does not have OOR (original object return), it constructs a new,
        # equivalent object each time you retrieve an attribute.
        # So we can free right after we create our list
        agent_lib.call(
            lib_contrast.free_eval_result,
            is_valid_return,
            results,
        )


def map_result_and_free_check_query_sink_result(
    ret: int,
    results: ctypes._Pointer[agent_lib.CCheckQuerySinkResult],
    user_input: str,
    input_index: int,
    input_len: int,
    is_valid_return: Callable[[int], bool],
) -> InjectionResult | None:
    if ret == 0 and bool(results):
        evaluation = InjectionResult(
            user_input, input_index, input_len, results.contents
        )

        # ctypes does not have OOR (original object return), it constructs a new,
        # equivalent object each time you retrieve an attribute.
        # So we can free right after we create our list
        agent_lib.call(
            lib_contrast.free_check_query_sink_result,
            is_valid_return,
            results,
        )

        return evaluation
    return None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.triggers.session_rewriting_rule import (
    SessionRewritingRule,
)
from contrast.agent.policy.registry import register_trigger_rule


register_trigger_rule(
    SessionRewritingRule.from_nodes(
        "session-rewriting",
        [
            {
                # default use_cookies is True which is a safe value
                "module": "beaker.session",
                "class_name": "Session",
                "method_name": "__init__",
                "source": "ARG_3,KWARG:use_cookies",
            }
        ],
    )
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import os
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")

ENV_PREFIX = "CONTRAST"


def env_name(name):
    return "__".join([x for x in [ENV_PREFIX, name.replace(".", "__")] if x]).upper()


class ConfigBuilder:
    def __init__(self):
        self.default_options = []

    def build(self, config: dict, yaml_config: list[tuple[dict, str]]):
        """
        Given a dict config, iterate over the default_options and check if
        the corresponding config key/value should be either :
        1. replaced by an existing env var
        2. keep existing config key/val but type-cast the value
        3. add a new key/default_value to the config

        :param config: dict config
        :return: str if error was set, config dict is updated pass by reference
        """
        error = None
        for option in self.default_options:
            option_name = option.canonical_name
            type_cast = option.type_cast
            env_key = env_name(option_name)
            env_override = os.environ.get(env_key)
            if env_override is not None:
                try:
                    option.env_value = type_cast(env_override)
                    option.name = env_key
                except Exception as e:
                    logger.exception("Failed to initialize config")
                    if error is None:
                        error = f"Invalid value on {option.canonical_name} - {e}"
            for cfg, filename in yaml_config:
                file_override = cfg.get(option_name, None)
                if file_override is not None:
                    if not option.file_values and not option.file_sources:
                        option.file_values = []
                        option.file_sources = []
                    try:
                        option.file_values.append(type_cast(file_override))
                        option.file_sources.append(filename)
                    except Exception as e:
                        logger.exception("Failed to initialize config")
                        if error is None:
                            error = f"Invalid value on {option_name} - {e}"
            config[option_name] = option
        return error



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements patches for the operator module

Our propagation rewrites are implemented in terms of these patches, so they
must always be enabled when Assess is enabled.
"""

import sys
import contrast

from contrast.agent.assess.policy.propagators.base_propagator import SUPPORTED_TYPES

from contrast.agent import scope
from contrast.agent.assess.utils import is_trackable
from contrast.agent.policy import patch_manager
from contrast.patches.utils import analyze_policy, get_propagation_node
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)
from contrast.agent.assess.policy.propagators import BasePropagator, PROPAGATOR_ACTIONS
from contrast.agent.assess.policy.preshift import Preshift

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_mod_propagator(policy_node):
    """
    Build propagator function used in the operator.mod patch.

    Part of the complication here is that the format propagator expects the format
    string to be `OBJ` (not `ARG_0`). We need to perform some manual rearrangement of
    the operator.mod arguments for format propagation to work correctly.

    TODO: PYT-3718 - In the future, the format propagator should be able to select the
    format string out of a preshift object using the policy node (instead of being
    hardcoded to extract the format string from OBJ).
    """
    propagator_class = PROPAGATOR_ACTIONS.get(policy_node.action, BasePropagator)

    @fail_quietly("failed to propagate through mod")
    @scope.contrast_scope()
    @scope.propagation_scope()
    def propagate_mod(ret, orig_args, kwargs):
        assert kwargs == {}  # mod takes no kwargs
        fmt_str, args = orig_args
        if not isinstance(fmt_str, bytearray) and ret is fmt_str:
            return

        context = contrast.REQUEST_CONTEXT.get()
        if context is None or context.stop_propagation:
            return

        if not isinstance(args, tuple):
            args = (args,)

        preshift = Preshift(fmt_str, args, {})

        propagator = propagator_class(policy_node, preshift, ret)
        if not propagator.needs_propagation:
            return

        propagator.track_and_propagate(ret)

        logger.debug("Propagator %s found: propagated to %s", policy_node.name, id(ret))

        context.propagated()

    return propagate_mod


def build_add_hook(original_func, patch_policy):
    policy_node = get_propagation_node(patch_policy)

    def add(wrapped, instance, args, kwargs):
        del instance

        if not isinstance(args[0], SUPPORTED_TYPES) and not isinstance(
            args[1], SUPPORTED_TYPES
        ):
            # This is an operation we don't support, so we return the result
            # without any tracking.
            # Don't enter scope, because this could be an operation against
            # a class with a custom __add__ method, and we want to track
            # dataflow through that user code.
            return wrapped(*args, **kwargs)

        with scope.contrast_scope():
            result = wrapped(*args, **kwargs)
        if not is_trackable(result) or scope.in_contrast_or_propagation_scope():
            return result

        analyze_policy(policy_node.name, result, args, kwargs)

        return result

    return wrap_and_watermark(original_func, add)


def build_mod_hook(original_func, patch_policy):
    propagation_func = build_mod_propagator(patch_policy.propagator_nodes[0])

    def mod(wrapped, instance, args, kwargs):
        del instance

        format_string = args[0]
        if not isinstance(format_string, SUPPORTED_TYPES):
            # This is a modulo operation that we don't support,
            # so we return the result without any tracking.
            # Don't enter scope, because this could be an operation
            # against a class with a custom __mod__ method, and we
            # want to track dataflow through that user code.
            return wrapped(*args, **kwargs)

        with scope.contrast_scope():
            result = wrapped(*args, **kwargs)
        if not is_trackable(result) or scope.in_contrast_or_propagation_scope():
            return result
        propagation_func(result, args, kwargs)

        return result

    return wrap_and_watermark(original_func, mod)


def patch_operator(operator_module):
    build_and_apply_patch(operator_module, "add", build_add_hook)
    build_and_apply_patch(operator_module, "iadd", build_add_hook)
    build_and_apply_patch(operator_module, "mod", build_mod_hook)


def register_patches():
    register_module_patcher(patch_operator, "operator")


def reverse_patches():
    unregister_module_patcher("operator")
    operator_module = sys.modules.get("operator")
    if not operator_module:
        return

    patch_manager.reverse_patches_by_owner(operator_module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.


class SecureFlagRuleMixin:
    @property
    def name(self):
        return "secure-flag-missing"

    def is_violated_properties(*args, **kwargs):
        """Override for safety. In theory, this should never be called."""
        return False

    def is_violated(self, value):
        """Rule is violated if config value is missing (None) or set to False"""
        if self.count_threshold_reached():
            return False
        return not value



# The inspect.formatargspec() function was dropped in Python 3.11 but we need
# need it for when constructing signature changing decorators based on result of
# inspect.getargspec() or inspect.getfullargspec(). The code here implements
# inspect.formatargspec() base on Parameter and Signature from inspect module,
# which were added in Python 3.6. Thanks to Cyril Jouve for the implementation.

try:
    from inspect import Parameter, Signature
except ImportError:
    from inspect import formatargspec
else:
    def formatargspec(args, varargs=None, varkw=None, defaults=None,
                      kwonlyargs=(), kwonlydefaults={}, annotations={}):
        if kwonlydefaults is None:
            kwonlydefaults = {}
        ndefaults = len(defaults) if defaults else 0
        parameters = [
            Parameter(
                arg,
                Parameter.POSITIONAL_OR_KEYWORD,
                default=defaults[i] if i >= 0 else Parameter.empty,
                annotation=annotations.get(arg, Parameter.empty),
            ) for i, arg in enumerate(args, ndefaults - len(args))
        ]
        if varargs:
            parameters.append(Parameter(varargs, Parameter.VAR_POSITIONAL))
        parameters.extend(
            Parameter(
                kwonlyarg,
                Parameter.KEYWORD_ONLY,
                default=kwonlydefaults.get(kwonlyarg, Parameter.empty),
                annotation=annotations.get(kwonlyarg, Parameter.empty),
            ) for kwonlyarg in kwonlyargs
        )
        if varkw:
            parameters.append(Parameter(varkw, Parameter.VAR_KEYWORD))
        return_annotation = annotations.get('return', Signature.empty)
        return str(Signature(parameters, return_annotation=return_annotation))


# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import functools
import inspect
from inspect import iscoroutinefunction
from types import ModuleType


from contrast.utils.patch_utils import register_module_patcher
from contrast.agent.policy import patch_manager, registry
from contrast.agent.protect.policy import apply_protect_patch
from contrast.agent.assess.policy.patches import (
    build_assess_method,
    build_assess_method_legacy,
    build_assess_classmethod,
    build_assess_deadzone,
    build_assess_property_fget,
    apply_cached_property,
    build_assess_async_method,
    build_assess_async_deadzone,
)
from contrast.utils.patch_utils import repatch_module
from contrast.utils.safe_import import safe_import


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


# Maintains a record of all patch locations that have been applied
PATCH_LOCATIONS = set()


@functools.lru_cache(maxsize=1)
def _get_werkzeug_cached_property():
    return safe_import("werkzeug.utils.cached_property")


def remove_patch_location(owner, name):
    """
    Remove patch located at owner (class or module) and name
    from patch locations.
    """
    if inspect.ismodule(owner):
        policy_patch_name = f"{owner.__name__}.{name}"
    elif inspect.isclass(owner):
        policy_patch_name = f"{owner.__module__}.{owner.__name__}.{name}"

    if policy_patch_name in PATCH_LOCATIONS:
        PATCH_LOCATIONS.remove(policy_patch_name)


def apply_assess_patch(patch_site, patch_policy):
    """
    Store the original method implementation under our custom "cs_assess_{name}" so we
    can call the old method from the new method in order to return the same result

    :param cls_or_module: Class or module to patch
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :param methods: methods in cls_or_module
    :param node_method: tuple of string and actual method
        ('get_raw_uri', <unbound method django.http.request.HttpRequest.get_raw_uri>)

    Static methods and class methods are implemented as descriptors. When one of these
    methods is called on the parent object, it uses the descriptor protocol, which
    means that the underlying function objects are not seen directly by the caller.

    In order to access the staticmethod/classmethod objects directly, they must be
    accessed via the parent object's __dict__ attribute.

    For example,

    class D:
        @staticmethod
        def sm():
            pass

    If I do this:
    D.sm will call __get__ under the hood to get the function definition (this is
    what's wrapped inside the staticmethod or classmethod obj in this case)

    If we do this:
    D.__dict__['sm'] we get the underlying staticmethod/classmethod object. This is
    required because we need to do a type check in order to replace the original obj
    with our own function of the correct type.

    This shows examples of C code re-written in python for the __get__ desc
    implementation for classmethod, staticmethod and function
    https://docs.python.org/2.7/howto/descriptor.html#functions-and-methods

    For additional details on PY3:
    https://docs.python.org/3/howto/descriptor.html#functions-and-methods
    """
    logger.debug("Applying assess patch to %s", patch_policy.name)

    method_name = patch_policy.method_name
    has_key = hasattr(patch_site, "__dict__") and method_name in patch_site.__dict__

    static_method = has_key and isinstance(
        patch_site.__dict__[method_name], staticmethod
    )
    class_method = has_key and isinstance(patch_site.__dict__[method_name], classmethod)

    old_method = _get_original_method(
        patch_policy, patch_site, method_name, static_method, class_method
    )

    werkzeug_cached_property = _get_werkzeug_cached_property()
    if werkzeug_cached_property is not None and isinstance(
        old_method, werkzeug_cached_property
    ):
        # return so that we don't prematurely run policy analysis.
        return apply_cached_property(patch_site, patch_policy, method_name, old_method)
    elif isinstance(old_method, property):
        fget = build_assess_property_fget(old_method, patch_policy)
        new_method = property(fget=fget, fset=old_method.fset, fdel=old_method.fdel)
    elif class_method:
        new_method = classmethod(build_assess_classmethod(old_method, patch_policy))
    elif static_method:
        new_method = staticmethod(build_assess_classmethod(old_method, patch_policy))
    elif iscoroutinefunction(old_method):
        new_method = (
            build_assess_async_deadzone(old_method, patch_policy)
            if patch_policy.is_deadzone
            else build_assess_async_method(old_method, patch_policy)
        )
    elif patch_policy.module in ("falcon.util.uri"):
        new_method = build_assess_method_legacy(old_method, patch_policy)
    else:
        # We only support deadzones for methods/functions right now.
        if patch_policy.is_deadzone and not patch_policy.deadzone_enabled:
            return False

        new_method = (
            build_assess_deadzone(old_method, patch_policy)
            if patch_policy.is_deadzone
            else build_assess_method(old_method, patch_policy)
        )

    # OPTIMIZATION: If we know this patch location is only one type (trigger,
    # propagation, source) and only one node, then we can assign that analysis code here
    # at patch-time and save time when patch is called. Most patches will benefit
    # from optimization as very few functions are more than one thing (example:
    # re.split is both a trigger for redos rule and a propagator).
    if len(patch_policy.all_nodes) == 1:
        patch_policy.assign_analysis_func()

    try:
        patch_manager.patch(patch_site, method_name, new_method)
        logger.debug("added patch to %s.%s", patch_site, method_name)
    except Exception as e:
        logger.debug(
            "unable to patch %s method of %s: %s", method_name, patch_site, str(e)
        )
        return False

    return True


def _get_original_method(
    patch_policy, patch_site, method_name, static_method, class_method
):
    # Need to make sure we get original static/class function
    if static_method or class_method:
        old_method = patch_site.__dict__[method_name]
    else:
        # get old function/property/method
        try:
            old_method = getattr(patch_site, patch_policy.method_name)
        except AttributeError:
            # Some python/framework versions or platforms will not have certain
            # methods/funcs defined so we skip them here. This isn't a failure
            # so logging is not necessary.
            return False

    return old_method


def apply_module_patches(module: ModuleType):
    """
    Apply patches to all methods and functions in a module as dictated by policy.

    This function might be called multiple times per module if multiple agent modes are
    enabled or in certain special cases (see direct references to this function). The
    global PATCH_LOCATIONS set ensures instrumentation is only actually applied once.
    """
    logger.debug("Applying module patches", module=module.__name__)

    _apply_v1_patches(module)
    _apply_v2_patches(module)

    # It's possible that the current module contains multiple references to the
    # function we replaced, but that only one of them is represented in policy. We do a
    # quick repatching pass over the current module here to make sure we cover all of
    # the references that may have been missed.
    repatch_module(module)

    # EDGE CASE PYT-1065: Werkzeug==0.16.x
    # This version of Werkzeug keeps a reference to the module in _real_module
    # which requires us to repatch functions in this second reference, too.
    if hasattr(module, "_real_module"):
        repatch_module(module._real_module)


def reverse_module_patches(module: ModuleType):
    """
    Reverse all patches applied to a module.
    """
    for patch_policy in registry.get_policy_by_module(module.__name__):
        if patch_policy.name in PATCH_LOCATIONS:
            PATCH_LOCATIONS.remove(patch_policy.name)
        if patch_policy.class_name:
            patch_manager.reverse_patches_by_owner(
                getattr(module, patch_policy.class_name)
            )

    patch_manager.reverse_patches_by_owner(module)


def _apply_v1_patches(module: ModuleType):
    module_policy = registry.get_policy_by_module(module.__name__)
    if module_policy is None:
        logger.debug("WARNING: No module policy found", module=module.__name__)
        return

    from contrast.agent import agent_state, patch_controller

    preinstrument = patch_controller.is_preinstrument_flag_set()

    for patch_policy in module_policy:
        if patch_policy.name in PATCH_LOCATIONS:
            continue

        # If the module has no policy nodes, or if none of the nodes are policy
        # patches, then there's nothing to do here.
        if not patch_policy.has_patches:
            continue

        # If nothing is on, there's nothing to do.
        if (
            not agent_state.module.protect_enabled
            and not agent_state.module.assess_enabled
            and not preinstrument
        ):
            continue

        # If only Protect is on and this isn't a protect patch, there's nothing to do.
        if (
            agent_state.module.protect_enabled
            and not agent_state.module.assess_enabled
            and not preinstrument
            and not patch_policy.is_protect_mode
        ):
            continue

        if patch_policy.class_name:
            patch_site = getattr(module, patch_policy.class_name, None)
            if patch_site is None:
                continue
        else:
            patch_site = module

        try:
            if (
                agent_state.module.protect_enabled or preinstrument
            ) and patch_policy.is_protect_mode:
                apply_protect_patch(patch_site, patch_policy)
        except Exception:
            logger.debug("Failed to apply protect patch for %s", patch_policy.name)

        try:
            if agent_state.module.assess_enabled or preinstrument:
                apply_assess_patch(patch_site, patch_policy)
        except Exception:
            logger.debug("Failed to apply assess patch for %s", patch_policy.name)

        PATCH_LOCATIONS.add(patch_policy.name)


@functools.cache
def _apply_v2_patches(module: ModuleType):
    from contrast.agent.policy import registry_v2

    patch_locations = {
        location
        for location in registry_v2.get_policy_locations()
        if location.module == module.__name__
    }
    if len(patch_locations) == 0:
        logger.debug("No v2 module policy found", module=module.__name__)
        return

    for location in patch_locations:
        patch_site = (
            module
            if location.class_name is None
            else getattr(module, location.class_name, None)
        )
        if patch_site is None:
            logger.debug(
                "Patch site not found", location=location, module=module.__name__
            )
            continue
        original_func = getattr(patch_site, location.method_name, None)
        if original_func is None:
            logger.debug(
                "Original function not found",
                method=location.method_name,
                module=module.__name__,
            )
            continue

        patch_manager.patch(
            patch_site,
            location.method_name,
            registry_v2.build_generic_contrast_wrapper(
                original_func, module_name=location.module
            ),
        )


def register_policy_patches(*, protect_mode: bool):
    """
    Use policy to register import hooks for each module that requires patches

    If protect_mode, patch only module patches with protect if trigger node has
    protect_mode: true.
    If not protect_mode, we will patch all module patches with assess.
    """
    from contrast.agent.policy import registry_v2

    modules_to_patch = set(registry.get_patch_policies(protect=protect_mode)) | {
        location.module for location in registry_v2.get_policy_locations()
    }

    for module_name in modules_to_patch:
        logger.debug("Registering import hook for %s", module_name)
        register_module_patcher(apply_module_patches, module_name)


def apply_patch_to_dynamic_property(class_to_patch, property_name, tags):
    """
    Take the property of a class we want to patch and:
        1. create a source node to store in policy
        2. patch the original property with our own code, including
            the policy instance with the new dynamic source.

    This means that the next time the cls.property is called,
    we will inject ourselves and run source policy.

    NOTE: adding the dynamic source to policy BEFORE patching is critical order,
    given that we have to patch with the policy instance that has this dynamic source.
    This could later be modified to not need this requirement.
    """
    module = class_to_patch.__module__
    class_name = class_to_patch.__name__

    patch_policy = registry.register_dynamic_source(
        module,
        class_name,
        property_name,
        tags,
        policy_patch=False,
    )

    orig_property = getattr(class_to_patch, property_name)

    def fset(cls_instance, value):
        # while good instinct would lead us to use setattr here instead of __dict__,
        # doing so does not work because we are in fact within a setter!
        cls_instance.__dict__[property_name] = value

    new_property = property(
        fget=build_assess_property_fget(orig_property, patch_policy),
        fset=fset,
        fdel=getattr(orig_property, "fdel", None),
    )

    patch_manager.patch(class_to_patch, property_name, new_property)

    return True



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.object_utils import get_name
from contrast.utils.decorators import fail_loudly

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def get_original_app_or_fail(app, orig_app_class):
    _app = find_original_application(app, orig_app_class)
    if isinstance(_app, orig_app_class):
        return _app

    msg = (
        f"Unable to find the original {get_name(orig_app_class)} Application object. "
        f"Please provide the original {get_name(orig_app_class)} Application object as the second argument to "
        "ContrastMiddleware."
    )

    logger.error(msg)
    raise RuntimeError(msg)


@fail_loudly("Unable to find original application object")
def find_original_application(app, orig_app_class, depth=20):
    """
    Recursively search through the middleware chain of `app` for an
    application of type `orig_app_class`. Most WSGI/ASGI middlewares are implemented as
    classes and maintain a reference to their wrapped app as an attribute. This function
    makes a best effort to find this attribute on each successive middleware until it
    sees an instance of the desired class.

    This method is not intended to succeed every time, as it is based off of several
    assumptions that will not always be true. It needs only succeed often enough that
    customers rarely need to supply additional information to framework-specific
    middleware constructors.

    This delegates to _find_original_application to ensure that we don't get an extra
    @fail_loudly on each recursive call.
    """
    return _find_original_application(app, orig_app_class, depth)


def _find_original_application(app, orig_app_class, depth):
    if isinstance(app, orig_app_class):
        return app
    if depth == 0:
        return None

    # this list is in approximate order of expected app attribute name; since this
    # algorithm is greedy, we iterate over the best candidates first
    for attr_name in [
        "app",
        "application",
        "wsgi_app",
        "asgi_app",
        "_app",
        "_wsgi_app",
        "_asgi_app",
        "_application",
        "wsgi_application",
        "asgi_application",
        "_wsgi_application",
        "_asgi_application",
        "wrapped_app",
        "wrapped_application",
        "wrapped",
        "wrap_app",
        "wrap",
        "registry",  # pyramid
        "_wrapped",
        "_wrapped_app",
        "_wrapped_application",
        "_wrap_app",
        "_wrap",
    ]:
        attr = getattr(app, attr_name, None)

        # The name check is a very unfortunate thing we have to do since
        # Pyramid's Registry obj is not a callable.
        if callable(attr) or attr.__class__.__name__ == "Registry":
            return _find_original_application(attr, orig_app_class, depth - 1)

    return None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import locale
import sys

DEFAULT_ENCODING = (
    locale.getencoding()
    if sys.version_info >= (3, 11)
    else locale.getpreferredencoding(do_setlocale=False)
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations

from contrast.agent.assess.rules.base_rule import BaseRule
from contrast.utils.timer import now_ms
from contrast_fireball import AssessEvent, AssessFinding, AssessRequest


CURRENT_FINDING_VERSION = 4


def build_finding(
    rule: BaseRule,
    properties: dict[str, str],
    events: list[AssessEvent] | None = None,
    request: AssessRequest | None = None,
    **kwargs,
) -> AssessFinding:
    """
    Builds a finding for the given rule. Properties are expected when the rule
    is a configuration or response scanning rule. Events are expected for
    dataflow rules. Request is expected for dataflow and response scanning rules.
    These expectations are from TeamServer, not checked in this function.
    """
    from contrast.agent import agent_state

    if events:
        kwargs.update(events=events)
    if request:
        kwargs.update(request=request)

    return AssessFinding(
        version=CURRENT_FINDING_VERSION,
        events=events or [],
        properties=properties,
        routes=[],
        rule_id=rule.name,
        tags=agent_state.module.assess_tags,
        request=request,
        evidence=None,
        hash=int(rule.generate_preflight_hash(**kwargs)),
        created=events[-1].time if events else now_ms(),
    )


def send_finding(finding: AssessFinding, context=None) -> None:
    """
    Send a finding by either appending it to request context OR sending it immediately.

    If `context` exists, the agent should not send the finding immediately because the
    current route needs to be appended (once it is available at the end of the request).
    """
    if context:
        if (
            context.input_exclusions_trigger_time
            and context.exclusions.evaluate_assess_trigger_time_exclusions(
                context, finding
            )
        ):
            return

        context.findings.append(finding)
        return

    from contrast.reporting import teamserver_messages
    from contrast.agent import agent_state

    agent_state.module.reporting_client.add_message(
        teamserver_messages.Preflight([finding], None)
    )



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from functools import cached_property
from collections.abc import Sequence
from typing import Callable
from contrast.agent.assess.policy import (
    propagation_policy,
    source_policy,
    trigger_policy,
)
from contrast.agent.assess.policy.deadzone_node import DeadZoneNode
from contrast.agent.assess.policy.propagation_node import PropagationNode
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.policy.policy_node import PolicyNode
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.settings import Settings


class PatchLocationPolicy:
    """
    Container for all policy nodes that pertain to a single patch location

    A patch location is a specific function or method that we patch. A single patch
    location may correspond to multiple policy nodes. Since import hooks are applied at
    the module level, there is a one-to-many relationship between each module and the
    corresponding PatchLocationPolicy instances.
    """

    def __init__(self, node: PolicyNode):
        self.module = node.module
        self.class_name = node.class_name
        self.method_name = node.method_name
        # This is a copy of the name attribute of each PolicyNode represented by this
        # PatchLocationPolicy instances. All policy nodes represented by this location
        # must have the same name.
        self.name = node.name

        self.deadzone_nodes: list[DeadZoneNode] = []
        self.source_nodes: list[SourceNode] = []
        self.propagator_nodes: list[PropagationNode] = []
        self.trigger_nodes: list[TriggerNode] = []
        self.analysis_func: Callable | None = None

    def add_node(self, node: PolicyNode):
        """
        Add a policy node for this patch location
        """
        # This is a sanity check to ensure that we don't accidentally add a node that
        # corresponds to a different patch location.
        assert node.name == self.name

        if isinstance(node, DeadZoneNode):
            self.deadzone_nodes.append(node)
        elif isinstance(node, PropagationNode):
            self.propagator_nodes.append(node)
        elif isinstance(node, SourceNode):
            self.source_nodes.append(node)
        elif isinstance(node, TriggerNode):
            self.trigger_nodes.append(node)

    def assign_analysis_func(self):
        """
        Store the analysis function as an optimization.
        This should only be called when we know there is only one node for this
        patch location and for assess only.
        """
        node = self.all_nodes[0]

        if isinstance(node, TriggerNode):
            self.analysis_func = trigger_policy.apply
        elif isinstance(node, PropagationNode):
            self.analysis_func = propagation_policy.apply
        elif isinstance(node, SourceNode):
            self.analysis_func = source_policy.apply

    @property
    def has_patches(self) -> bool:
        """
        Indicates whether any nodes at this location require policy patches
        """
        return any([node.policy_patch for node in self.all_nodes])

    @property
    def is_deadzone(self) -> bool:
        """
        Indicates if there are any deadzone nodes.

        NOTE: at this time, deadzone nodes and any other nodes are mutually exclusive.
        That is, all patches will receive the deadzone patch.
        """
        return bool(self.deadzone_nodes)

    @property
    def deadzone_enabled(self) -> bool:
        """
        This location's deadzone is enabled if at least one deadzone node exists and
        for ANY existing deadzone node, either:
        (a) it does not have an associated config option
        (b) its associated config option indicates that the deadzone should be applied
        """
        settings = Settings()

        return any(
            [
                node.config_option is None
                # common-config options must default to True. This means options must be
                # worded in such a way that setting to False will enable the deadzones
                or (
                    settings.config and settings.config.get(node.config_option) is False
                )
                for node in self.deadzone_nodes
            ]
        )

    @property
    def is_protect_mode(self) -> bool:
        """
        Check if any of the trigger nodes have protect_mode turned on.
        """
        if not self.trigger_nodes:
            return False

        return any([node.protect_mode for node in self.trigger_nodes])

    @cached_property
    def all_nodes(self) -> Sequence[PolicyNode]:
        """
        Returns all nodes in this policy as a list
        """
        return (
            self.source_nodes
            + self.propagator_nodes
            + self.trigger_nodes
            + self.deadzone_nodes
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.name})>"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.config.base_django_config_rule import (
    BaseDjangoConfigRule,
)
from contrast.agent.assess.rules.config.session_age_rule import SessionAgeRuleMixin


class DjangoSessionAgeRule(SessionAgeRuleMixin, BaseDjangoConfigRule):
    SETTINGS_VALUE = "SESSION_COOKIE_AGE"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from enum import Enum, auto


class Mode(Enum):
    OFF = auto()
    MONITOR = auto()
    BLOCK = auto()
    BLOCK_AT_PERIMETER = auto()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contextvars import Context
import sys
import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    register_module_patcher,
    unregister_module_patcher,
    wrap_and_watermark,
)

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_start_patch(orig_func, _):
    """
    Only used in python 3.13 and earlier.

    We store the parent thread's scope and request context on the thread object in
    `start` (which runs in the parent thread). Then, in `_bootstrap_inner`, we retrieve
    these values and set the corresponding contextvars in the child thread.

    In python 3.14, thread contextvar inheritance behavior was adjusted. 3.14+ uses
    different thread patching machinery.
    """

    def start(wrapped, instance, args, kwargs):
        context = contrast.REQUEST_CONTEXT.get()

        try:
            # Save the scope of the current active contextvars.Context to copy to the new thread
            instance.cs__parent_scope = scope.current_scope()
            instance.cs__parent_context = context
        except Exception:
            logger.exception("Failed to instrument thread start")

        return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_func, start)


def build_bootstrap_inner_patch(orig_func, _):
    """
    Only used in pythno 3.13 and earlier. See build_start_patch.
    """

    def _bootstrap_inner(wrapped, instance, args, kwargs):  # pragma: no cover
        # The new thread inherits the scope from the thread that created it
        try:
            scope.set_scope(*instance.cs__parent_scope)
        except Exception:
            logger.exception("Failed to initialize thread scope")

        with contrast.lifespan(instance.cs__parent_context):
            # Ensure child thread still runs with the same parent request context
            # even if the parent thread has already exited as long as
            # the parent thread is in request context.
            result = wrapped(*args, **kwargs)

        # We expect result to be None, but this is done for consistency/safety
        return result

    return wrap_and_watermark(orig_func, _bootstrap_inner)


def _set_contrast_contextvars(target_context: Context):
    """
    Set all contrast-relevant contextvars in the provided Context, taking their values
    from the current context. Overwrites them if they are already present in the
    provided context.
    """
    parent_scope = scope.current_scope()
    parent_context = contrast.REQUEST_CONTEXT.get()

    def set_vars():
        scope.set_scope(*parent_scope)
        contrast.REQUEST_CONTEXT.set(parent_context)

    target_context.run(set_vars)


def build_thread_init_patch(orig_func, _):
    """
    Only used in python 3.14 and later.

    Patch for threading.Thread.__init__. Executes in the parent thread. We need to make
    sure scope and request context are transferred to the child thread. This behavior is
    controlled by both sys.flags.thread_inherit_context and the `context` kwarg to this
    function.

    See https://docs.python.org/3.14/library/threading.html#threading.Thread.
    """

    def thread_init(wrapped, instance, args, kwargs):
        # `context` is kwarg-only. It defaults to None, so we don't need special
        # handling for the case where it is not provided at all.
        passed_context = kwargs.get("context")
        if passed_context is None:
            # sys.flags controls the behavior in this case
            if sys.flags.thread_inherit_context == 0:
                new_context = Context()
                _set_contrast_contextvars(new_context)
                kwargs["context"] = new_context
            # the `inherit_context != 0` case is a noop; contextvars are inherited
        else:
            _set_contrast_contextvars(passed_context)

        return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_func, thread_init)


def patch_threading(threading_module):
    if sys.version_info[:2] <= (3, 13):
        # py313 this can be removed once we no longer need to support 3.13
        build_and_apply_patch(
            threading_module.Thread,
            "start",
            build_start_patch,
        )
        # This instruments the method that actually runs inside the system thread
        build_and_apply_patch(
            threading_module.Thread,
            "_bootstrap_inner",
            build_bootstrap_inner_patch,
        )
    else:
        build_and_apply_patch(
            threading_module.Thread, "__init__", build_thread_init_patch
        )


def register_patches():
    register_module_patcher(patch_threading, "threading")


def reverse_patches():
    unregister_module_patcher("threading")
    threading = sys.modules.get("threading")
    if not threading:
        return

    patch_manager.reverse_patches_by_owner(threading.Thread)



"""An asyncio-based implementation of the file lock."""

from __future__ import annotations

import asyncio
import contextlib
import logging
import os
import time
from dataclasses import dataclass
from inspect import iscoroutinefunction
from threading import local
from typing import TYPE_CHECKING, Any, NoReturn, cast

from ._api import BaseFileLock, FileLockContext, FileLockMeta
from ._error import Timeout
from ._soft import SoftFileLock
from ._unix import UnixFileLock
from ._windows import WindowsFileLock

if TYPE_CHECKING:
    import sys
    from collections.abc import Callable
    from concurrent import futures
    from types import TracebackType

    if sys.version_info >= (3, 11):  # pragma: no cover (py311+)
        from typing import Self
    else:  # pragma: no cover (<py311)
        from typing_extensions import Self


_LOGGER = logging.getLogger("filelock")


@dataclass
class AsyncFileLockContext(FileLockContext):
    """A dataclass which holds the context for a ``BaseAsyncFileLock`` object."""

    #: Whether run in executor
    run_in_executor: bool = True

    #: The executor
    executor: futures.Executor | None = None

    #: The loop
    loop: asyncio.AbstractEventLoop | None = None


class AsyncThreadLocalFileContext(AsyncFileLockContext, local):
    """A thread local version of the ``FileLockContext`` class."""


class AsyncAcquireReturnProxy:
    """A context-aware object that will release the lock file when exiting."""

    def __init__(self, lock: BaseAsyncFileLock) -> None:  # noqa: D107
        self.lock = lock

    async def __aenter__(self) -> BaseAsyncFileLock:  # noqa: D105
        return self.lock

    async def __aexit__(  # noqa: D105
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        await self.lock.release()


class AsyncFileLockMeta(FileLockMeta):
    def __call__(  # type: ignore[override] # noqa: PLR0913
        cls,  # noqa: N805
        lock_file: str | os.PathLike[str],
        timeout: float = -1,
        mode: int = 0o644,
        thread_local: bool = False,  # noqa: FBT001, FBT002
        *,
        blocking: bool = True,
        is_singleton: bool = False,
        loop: asyncio.AbstractEventLoop | None = None,
        run_in_executor: bool = True,
        executor: futures.Executor | None = None,
    ) -> BaseAsyncFileLock:
        if thread_local and run_in_executor:
            msg = "run_in_executor is not supported when thread_local is True"
            raise ValueError(msg)
        instance = super().__call__(
            lock_file=lock_file,
            timeout=timeout,
            mode=mode,
            thread_local=thread_local,
            blocking=blocking,
            is_singleton=is_singleton,
            loop=loop,
            run_in_executor=run_in_executor,
            executor=executor,
        )
        return cast("BaseAsyncFileLock", instance)


class BaseAsyncFileLock(BaseFileLock, metaclass=AsyncFileLockMeta):
    """Base class for asynchronous file locks."""

    def __init__(  # noqa: PLR0913
        self,
        lock_file: str | os.PathLike[str],
        timeout: float = -1,
        mode: int = 0o644,
        thread_local: bool = False,  # noqa: FBT001, FBT002
        *,
        blocking: bool = True,
        is_singleton: bool = False,
        loop: asyncio.AbstractEventLoop | None = None,
        run_in_executor: bool = True,
        executor: futures.Executor | None = None,
    ) -> None:
        """
        Create a new lock object.

        :param lock_file: path to the file
        :param timeout: default timeout when acquiring the lock, in seconds. It will be used as fallback value in \
            the acquire method, if no timeout value (``None``) is given. If you want to disable the timeout, set it \
            to a negative value. A timeout of 0 means that there is exactly one attempt to acquire the file lock.
        :param mode: file permissions for the lockfile
        :param thread_local: Whether this object's internal context should be thread local or not. If this is set to \
            ``False`` then the lock will be reentrant across threads.
        :param blocking: whether the lock should be blocking or not
        :param is_singleton: If this is set to ``True`` then only one instance of this class will be created \
            per lock file. This is useful if you want to use the lock object for reentrant locking without needing \
            to pass the same object around.
        :param loop: The event loop to use. If not specified, the running event loop will be used.
        :param run_in_executor: If this is set to ``True`` then the lock will be acquired in an executor.
        :param executor: The executor to use. If not specified, the default executor will be used.

        """
        self._is_thread_local = thread_local
        self._is_singleton = is_singleton

        # Create the context. Note that external code should not work with the context directly and should instead use
        # properties of this class.
        kwargs: dict[str, Any] = {
            "lock_file": os.fspath(lock_file),
            "timeout": timeout,
            "mode": mode,
            "blocking": blocking,
            "loop": loop,
            "run_in_executor": run_in_executor,
            "executor": executor,
        }
        self._context: AsyncFileLockContext = (AsyncThreadLocalFileContext if thread_local else AsyncFileLockContext)(
            **kwargs
        )

    @property
    def run_in_executor(self) -> bool:
        """::return: whether run in executor."""
        return self._context.run_in_executor

    @property
    def executor(self) -> futures.Executor | None:
        """::return: the executor."""
        return self._context.executor

    @executor.setter
    def executor(self, value: futures.Executor | None) -> None:  # pragma: no cover
        """
        Change the executor.

        :param value: the new executor or ``None``
        :type value: futures.Executor | None

        """
        self._context.executor = value

    @property
    def loop(self) -> asyncio.AbstractEventLoop | None:
        """::return: the event loop."""
        return self._context.loop

    async def acquire(  # type: ignore[override]
        self,
        timeout: float | None = None,
        poll_interval: float = 0.05,
        *,
        blocking: bool | None = None,
    ) -> AsyncAcquireReturnProxy:
        """
        Try to acquire the file lock.

        :param timeout: maximum wait time for acquiring the lock, ``None`` means use the default
            :attr:`~BaseFileLock.timeout` is and if ``timeout < 0``, there is no timeout and
            this method will block until the lock could be acquired
        :param poll_interval: interval of trying to acquire the lock file
        :param blocking: defaults to True. If False, function will return immediately if it cannot obtain a lock on the
         first attempt. Otherwise, this method will block until the timeout expires or the lock is acquired.
        :raises Timeout: if fails to acquire lock within the timeout period
        :return: a context object that will unlock the file when the context is exited

        .. code-block:: python

            # You can use this method in the context manager (recommended)
            with lock.acquire():
                pass

            # Or use an equivalent try-finally construct:
            lock.acquire()
            try:
                pass
            finally:
                lock.release()

        """
        # Use the default timeout, if no timeout is provided.
        if timeout is None:
            timeout = self._context.timeout

        if blocking is None:
            blocking = self._context.blocking

        # Increment the number right at the beginning. We can still undo it, if something fails.
        self._context.lock_counter += 1

        lock_id = id(self)
        lock_filename = self.lock_file
        start_time = time.perf_counter()
        try:
            while True:
                if not self.is_locked:
                    _LOGGER.debug("Attempting to acquire lock %s on %s", lock_id, lock_filename)
                    await self._run_internal_method(self._acquire)
                if self.is_locked:
                    _LOGGER.debug("Lock %s acquired on %s", lock_id, lock_filename)
                    break
                if blocking is False:
                    _LOGGER.debug("Failed to immediately acquire lock %s on %s", lock_id, lock_filename)
                    raise Timeout(lock_filename)  # noqa: TRY301
                if 0 <= timeout < time.perf_counter() - start_time:
                    _LOGGER.debug("Timeout on acquiring lock %s on %s", lock_id, lock_filename)
                    raise Timeout(lock_filename)  # noqa: TRY301
                msg = "Lock %s not acquired on %s, waiting %s seconds ..."
                _LOGGER.debug(msg, lock_id, lock_filename, poll_interval)
                await asyncio.sleep(poll_interval)
        except BaseException:  # Something did go wrong, so decrement the counter.
            self._context.lock_counter = max(0, self._context.lock_counter - 1)
            raise
        return AsyncAcquireReturnProxy(lock=self)

    async def release(self, force: bool = False) -> None:  # type: ignore[override]  # noqa: FBT001, FBT002
        """
        Releases the file lock. Please note, that the lock is only completely released, if the lock counter is 0.
        Also note, that the lock file itself is not automatically deleted.

        :param force: If true, the lock counter is ignored and the lock is released in every case/

        """
        if self.is_locked:
            self._context.lock_counter -= 1

            if self._context.lock_counter == 0 or force:
                lock_id, lock_filename = id(self), self.lock_file

                _LOGGER.debug("Attempting to release lock %s on %s", lock_id, lock_filename)
                await self._run_internal_method(self._release)
                self._context.lock_counter = 0
                _LOGGER.debug("Lock %s released on %s", lock_id, lock_filename)

    async def _run_internal_method(self, method: Callable[[], Any]) -> None:
        if iscoroutinefunction(method):
            await method()
        elif self.run_in_executor:
            loop = self.loop or asyncio.get_running_loop()
            await loop.run_in_executor(self.executor, method)
        else:
            method()

    def __enter__(self) -> NoReturn:
        """
        Replace old __enter__ method to avoid using it.

        NOTE: DO NOT USE `with` FOR ASYNCIO LOCKS, USE `async with` INSTEAD.

        :return: none
        :rtype: NoReturn
        """
        msg = "Do not use `with` for asyncio locks, use `async with` instead."
        raise NotImplementedError(msg)

    async def __aenter__(self) -> Self:
        """
        Acquire the lock.

        :return: the lock object

        """
        await self.acquire()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """
        Release the lock.

        :param exc_type: the exception type if raised
        :param exc_value: the exception value if raised
        :param traceback: the exception traceback if raised

        """
        await self.release()

    def __del__(self) -> None:
        """Called when the lock object is deleted."""
        with contextlib.suppress(RuntimeError):
            loop = self.loop or asyncio.get_running_loop()
            if not loop.is_running():  # pragma: no cover
                loop.run_until_complete(self.release(force=True))
            else:
                loop.create_task(self.release(force=True))


class AsyncSoftFileLock(SoftFileLock, BaseAsyncFileLock):
    """Simply watches the existence of the lock file."""


class AsyncUnixFileLock(UnixFileLock, BaseAsyncFileLock):
    """Uses the :func:`fcntl.flock` to hard lock the lock file on unix systems."""


class AsyncWindowsFileLock(WindowsFileLock, BaseAsyncFileLock):
    """Uses the :func:`msvcrt.locking` to hard lock the lock file on windows systems."""


__all__ = [
    "AsyncAcquireReturnProxy",
    "AsyncSoftFileLock",
    "AsyncUnixFileLock",
    "AsyncWindowsFileLock",
    "BaseAsyncFileLock",
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import io

from contrast.agent.assess.utils import is_tracked
from contrast.utils.decorators import fail_quietly
from contrast.utils.assess.duck_utils import safe_getattr, safe_iterator

from contrast.utils.safe_import import safe_import
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

SUPPORTED_TYPES = (
    str,
    bytes,
    bytearray,
)

DJANGO_PROMISE_TYPE = safe_import("django.utils.functional.Promise")


@fail_quietly("Failed check if obj is tracked", return_value=False)
def recursive_is_tracked(obj):  # pylint: disable=too-many-return-statements
    """
    This is a fairly hot code path in some applications; be mindful of performance
    """
    if obj is None:
        return False

    if isinstance(obj, SUPPORTED_TYPES):
        return is_tracked(obj)

    if isinstance(obj, dict):
        return any(
            recursive_is_tracked(key) or recursive_is_tracked(value)
            for key, value in obj.items()
        )
    if isinstance(obj, io.IOBase):
        return safe_getattr(obj, "cs__tracked", False) or safe_getattr(
            obj, "cs__source", False
        )
    # These are the only safe objects where we can guarantee that we won't
    # destructively modify the object such that the application may see
    # different results.
    if isinstance(obj, (list, tuple)):
        return any(recursive_is_tracked(item) for item in safe_iterator(obj))

    if DJANGO_PROMISE_TYPE and isinstance(obj, DJANGO_PROMISE_TYPE):
        return is_tracked(str(obj))

    return is_tracked(obj)
